@@ -2,7 +2,7 @@
 /** @file SciTEBase.h
  ** Definition of platform independent base class of editor.
  **/
-// Copyright 1998-2002 by Neil Hodgson <neilh@scintilla.org>
+// Copyright 1998-2003 by Neil Hodgson <neilh@scintilla.org>
 // The License.txt file describes the conditions under which this software may be distributed.
 
 extern const char appName[];
@@ -19,11 +19,18 @@ extern const char fileWrite[];
 extern const char menuAccessIndicator[];
 
 #ifdef unix
+#include <limits.h>
+#ifdef PATH_MAX
+#define MAX_PATH PATH_MAX
+#else
 #define MAX_PATH 260
 #endif
+#endif
+
 #ifdef __vms
 const char *VMSToUnixStyle(const char *fileName);
 #endif
+
 #ifdef WIN32
 #ifdef _MSC_VER
 // Shut up level 4 warning:
@@ -31,6 +38,9 @@ const char *VMSToUnixStyle(const char *fileName);
 // warning C4800: forcing value to bool 'true' or 'false' (performance warning)
 #pragma warning(disable: 4710 4800)
 #endif
+#ifdef __DMC__
+#include <time.h>
+#endif
 #endif
 
 /**
@@ -99,12 +109,14 @@ public:
 };
 
 class PropSetFile : public PropSet {
+	bool lowerKeys;
 public:
-	PropSetFile();
+	PropSetFile(bool lowerKeys_=false);
 	~PropSetFile();
 	bool ReadLine(const char *data, bool ifIsTrue, const char *directoryForImports, SString imports[] = 0, int sizeImports = 0);
 	void ReadFromMemory(const char *data, int len, const char *directoryForImports, SString imports[] = 0, int sizeImports = 0);
 	bool Read(const char *filename, const char *directoryForImports, SString imports[] = 0, int sizeImports = 0);
+	void SetInteger(const char *key, sptr_t i);
 };
 
 class FilePath {
@@ -142,52 +154,190 @@ public:
 	}
 };
 
+// Related to Utf8_16::encodingType but with additional values at end
+enum UniMode {
+	uni8Bit=0, uni16BE=1, uni16LE=2, uniUTF8=3,
+	uniCookie=4
+};
+
+// State of folding in a given document, remembers line/state pairs,
+// restore aborts when these pairs mismatch
+class FoldState {
+private:
+	int *lines;
+	bool *folded;
+	int size;
+	int fill;
+	int readout;
+
+	void CopyFrom(const FoldState& b) {
+		Alloc(b.size);
+		memcpy(lines, b.lines, size*sizeof(int));
+		memcpy(folded, b.folded, size*sizeof(bool));
+		fill = b.fill;
+
+		readout = 0;
+	}
+
+public:
+	FoldState() {
+		lines = 0;
+		folded = 0;
+		size = 0;
+		fill = 0;
+		readout = -1;
+	}
+
+	FoldState &operator=(const FoldState &b) {
+		if (this != &b) {
+			CopyFrom(b);
+		}
+		return *this;
+	}
+
+	FoldState(const FoldState &b) {
+		lines = 0;
+		folded = 0;
+		readout = -1;
+		size = 0;
+		fill = 0;
+
+		CopyFrom(b);
+	}
+
+	void Alloc(int s) {
+		//assert(s>0);
+		//assert(size==0);
+
+		delete []lines;
+		lines = new int[s];
+		delete []folded;
+		folded = new bool[s];
+		size = s;
+		fill = 0;
+
+		//assert(lines && folded && size>0);
+	}
+
+	void Clear() {
+
+		delete []lines;
+		lines = 0;
+		delete []folded;
+		folded = 0;
+
+		size = 0;
+		fill = 0;
+		readout = -1;
+	}
+
+	virtual ~FoldState() {
+		Clear();
+	}
+
+	void PushState(int line, bool folded_) {
+		//assert(fill<size);
+		lines[fill] = line;
+		folded[fill] = folded_;
+		fill++;
+	}
+
+	// returns maximum numbers a readout can succeed
+	int BeginIteration() {
+		// strong assertion
+		//assert(fill==size);
+
+		readout = 0;
+		return fill;
+	}
+
+	// returns false at end of iteration
+	// no results are written then
+	bool GetState(int *pline, bool *pfold) {
+		//assert(readout>0);
+		if (readout >= fill) {
+			readout = -1;
+			return false;
+		} else {
+			if (pline && pfold) {
+				*pline = lines[readout];
+				*pfold = folded[readout];
+			}
+			readout++;
+			return true;
+		}
+	}
+};
+
 class Buffer : public RecentFile {
 public:
 	sptr_t doc;
 	bool isDirty;
 	bool useMonoFont;
+	UniMode unicodeMode;
 	time_t fileModTime;
 	SString overrideExtension;
+	FoldState foldState;
 	Buffer() :
-		RecentFile(), doc(0), isDirty(false), useMonoFont(false), fileModTime(0) {
+		RecentFile(), doc(0), isDirty(false), useMonoFont(false),
+		unicodeMode(uni8Bit), fileModTime(0), foldState() {
 	}
 
 	void Init() {
 		RecentFile::Init();
 		isDirty = false;
 		useMonoFont = false;
+		unicodeMode = uni8Bit;
 		fileModTime = 0;
 		overrideExtension = "";
+		foldState.Clear();
 	}
 };
 
 class BufferList {
+protected:
+	int current;
 public:
 	Buffer *buffers;
 	int size;
 	int length;
-	int current;
 	BufferList();
 	~BufferList();
 	void Allocate(int maxSize);
 	int Add();
 	int GetDocumentByName(const char *filename);
 	void RemoveCurrent();
+	int Current();
+	void SetCurrent(int index);
 };
 
 enum JobSubsystem {
     jobCLI = 0, jobGUI = 1, jobShell = 2, jobExtension = 3, jobHelp = 4, jobOtherHelp = 5};
+
+enum JobFlags {
+    jobForceQueue=1, jobHasInput=2, jobQuiet=4, // 8 reserved for jobVeryQuiet
+    jobRepSelMask=48, jobRepSelYes=16, jobRepSelAuto=32,
+	jobGroupUndo=64};
+
 class Job {
 public:
 	SString command;
 	SString directory;
+	SString input;
 	JobSubsystem jobType;
+	int flags;
 
 	Job();
 	void Clear();
 };
 
+// class to hold user defined keyboard short cuts
+class ShortcutItem {
+public:
+	SString menuKey; // the keyboard short cut
+	SString menuCommand; // the menu command to be passed to "SciTEBase::MenuCommand"
+};
+
 class LanguageMenuItem {
 public:
 	SString menuItem;
@@ -226,12 +376,15 @@ enum IndentationStatus {
 	isKeyWordStart	// Keywords that cause indentation
 };
 
+int IntFromHexDigit(const char ch);
+int IntFromHexByte(const char *hexByte);
+
 class StyleDefinition {
 public:
 	SString font;
 	int size;
-	ColourDesired fore;
-	ColourDesired back;
+	SString fore;
+	SString back;
 	bool bold;
 	bool italics;
 	bool eolfilled;
@@ -243,6 +396,9 @@ public:
 	             sdBold = 0x10, sdItalics = 0x20, sdEOLFilled = 0x40, sdUnderlined = 0x80,
 	              sdCaseForce = 0x100, sdVisible = 0x200, sdChangeable = 0x400} specified;
 	StyleDefinition(const char *definition);
+	bool ParseStyleDefinition(const char *definition);
+	long ForeAsLong();
+	long BackAsLong();
 };
 
 struct StyleAndWords {
@@ -271,13 +427,14 @@ protected:
 	RecentFile recentFileStack[fileStackMax];
 	enum { fileStackCmdID = IDM_MRUFILE, bufferCmdID = IDM_BUFFER };
 
-	enum { importMax = 20 };
+	enum { importMax = 50 };
 	SString importFiles[importMax];
 	enum { importCmdID = IDM_IMPORT };
 
 	SString findWhat;
 	SString replaceWhat;
 	Window wFindReplace;
+	Window wFindIncrement;
 	bool replacing;
 	bool havefound;
 	bool matchCase;
@@ -293,21 +450,26 @@ protected:
 	enum { maxParam = 4 };
 	Window wParameters;
 	SString parameterisedCommand;
+	char abbrevInsert[200];
 
 	enum { languageCmdID = IDM_LANGUAGE };
 	LanguageMenuItem *languageMenu;
 	int languageItems;
 
+	// an array of short cut items that are defined by the user in the properties file.
+	ShortcutItem *shortCutItemList; // array
+	int shortCutItems; // length of array
+
 	int codePage;
 	int characterSet;
+	UniMode unicodeMode;
 	SString language;
 	int lexLanguage;
 	SString overrideExtension;	///< User has chosen to use a particular language
-	enum {numWordLists = 6};
 	WordList apis;
+	SString apisFileNames;
 	SString functionDefinition;
 
-	int indentSize;
 	bool indentOpening;
 	bool indentClosing;
 	bool indentMaintain;
@@ -326,6 +488,7 @@ protected:
 	Window wContent;    ///< Contains wEditor and wOutput
 	Window wEditor;
 	Window wOutput;
+	Window wIncrement;
 #if PLAT_GTK
 	Window wDivider;	// Not used on Windows
 #endif
@@ -349,13 +512,16 @@ protected:
 	int visHeightStatus;
 	int visHeightEditor;
 	int heightBar;
+	// Prevent automatic load dialog appearing at the same time as
+	// other dialogs as this can leads to reentry errors.
 	int dialogsOnScreen;
 	bool topMost;
 	bool wrap;
 	bool wrapOutput;
-	bool checkIfOpen;
+	bool isReadOnly;
+	bool openFilesHere;
 	bool fullScreen;
-	enum { toolMax = 10 };
+	enum { toolMax = 50 };
 	Extension *extender;
 
 	int heightOutput;
@@ -367,6 +533,7 @@ protected:
 	bool localisationRead;
 	bool splitVertical;	///< @c true if the split bar between editor and output is vertical.
 	bool bufferedDraw;
+	bool twoPhaseDraw;
 	bool bracesCheck;
 	bool bracesSloppy;
 	int bracesStyle;
@@ -380,7 +547,14 @@ protected:
 	SString calltipWordCharacters;
 	SString calltipEndDefinition;
 	SString autoCompleteStartCharacters;
+	SString autoCompleteFillUpCharacters;
 	SString wordCharacters;
+	SString whitespaceCharacters;
+	int startCalltipWord;
+	int currentCallTip;
+	int maxCallTips;
+	SString currentCallTipWord;
+	int lastPosCallTip;
 
 	bool margin;
 	int marginWidth;
@@ -392,7 +566,8 @@ protected:
 
 	bool lineNumbers;
 	int lineNumbersWidth;
-	enum { lineNumbersWidthDefault = 40};
+	enum { lineNumbersWidthDefault = 4 };
+	bool lineNumbersExpand;
 
 	bool usePalette;
 	bool clearBeforeExecute;
@@ -424,9 +599,9 @@ protected:
 
 	PropSetFile propsUI;
 
-	PropSet propsStatus;
+	PropSetFile propsStatus;	// Not attached to a file but need SetInteger method.
 
-	enum { bufferMax = 10 };
+	enum { bufferMax = 100 };
 	BufferList buffers;
 
 	// Handle buffers
@@ -448,13 +623,16 @@ protected:
 	sptr_t SendEditor(unsigned int msg, uptr_t wParam = 0, sptr_t lParam = 0);
 	sptr_t SendEditorString(unsigned int msg, uptr_t wParam, const char *s);
 	sptr_t SendOutput(unsigned int msg, uptr_t wParam = 0, sptr_t lParam = 0);
+	sptr_t SendOutputString(unsigned int msg, uptr_t wParam, const char *s);
 	sptr_t SendFocused(unsigned int msg, uptr_t wParam = 0, sptr_t lParam = 0);
+	sptr_t SendPane(int destination, unsigned int msg, uptr_t wParam = 0, sptr_t lParam = 0);
 	sptr_t SendWindow(Window &w, unsigned int msg, uptr_t wParam=0, sptr_t lParam=0);
 	void SendChildren(unsigned int msg, uptr_t wParam = 0, sptr_t lParam = 0);
 	sptr_t SendOutputEx(unsigned int msg, uptr_t wParam = 0, sptr_t lParam = 0, bool direct = true);
 	int LengthDocument();
 	int GetCaretInLine();
 	void GetLine(char *text, int sizeText, int line=-1);
+	SString GetLine(int line=-1);
 	void GetRange(Window &win, int start, int end, char *text);
 	int IsLinePreprocessorCondition(char *line);
 	bool FindMatchingPreprocessorCondition(int &curLine, int direction, int condEnd1, int condEnd2);
@@ -471,86 +649,115 @@ protected:
 	void SaveRecentStack();
 	void LoadSession(const char *sessionName);
 	void SaveSession(const char *sessionName);
+	void SetIndentSettings();
+	void SetEol();
 	void New();
-	void Close(bool updateUI = true);
+	void Close(bool updateUI = true, bool loadingSession = false);
 	bool IsAbsolutePath(const char *path);
 	bool Exists(const char *dir, const char *path, char *testPath);
-	void OpenFile(bool initialCmdLine);
+	void OpenFile(bool suppressMessage);
 	virtual void OpenUriList(const char *) {};
 	virtual void AbsolutePath(char *fullPath, const char *basePath, int size) = 0;
 	virtual void FixFilePath();
-	virtual bool OpenDialog() = 0;
+	virtual bool OpenDialog(const char *filter=NULL) = 0;
 	virtual bool SaveAsDialog() = 0;
 	virtual void LoadSessionDialog() { };
 	virtual void SaveSessionDialog() { };
 	void CountLineEnds(int &linesCR, int &linesLF, int &linesCRLF);
-	bool Open(const char *file = 0, bool initialCmdLine = false,
-		bool forceLoad = false, bool maySaveIfDirty=true);
-	void OpenMultiple(const char *files = 0, bool initialCmdLine = false, bool forceLoad = false);
-	void OpenSelected();
+	enum OpenFlags {
+		ofNone=0, 		// Default
+		ofNoSaveIfDirty=1, 	// Suppress check for unsaved changes
+		ofForceLoad=2,	// Reload file even if already in a buffer
+		ofPreserveUndo=4,	// Do not delete undo history
+		ofQuiet=8		// Avoid "Could not open file" message
+	};
+	virtual bool PreOpenCheck(const char *file);
+	bool Open(const char *file, OpenFlags of=ofNone);
+	bool OpenSelected();
 	void Revert();
 	int SaveIfUnsure(bool forceQuestion = false);
 	int SaveIfUnsureAll(bool forceQuestion = false);
 	int SaveIfUnsureForBuilt();
 	bool Save();
-	virtual bool SaveAs(char *file = 0);
+	virtual bool SaveAs(const char *file = 0);
+	virtual void SaveACopy() = 0;
 	void SaveToHTML(const char *saveName);
+	void StripTrailingSpaces();
+	void EnsureFinalNewLine();
+	bool SaveBuffer(const char *saveName);
 	virtual void SaveAsHTML() = 0;
 	void SaveToRTF(const char *saveName, int start = 0, int end = -1);
 	virtual void SaveAsRTF() = 0;
 	void SaveToPDF(const char *saveName);
 	virtual void SaveAsPDF() = 0;
+	void SaveToTEX(const char *saveName);
+	virtual void SaveAsTEX() = 0;
+	void SaveToXML(const char *saveName);
+	virtual void SaveAsXML() = 0;
 	virtual void GetDefaultDirectory(char *directory, size_t size) = 0;
 	virtual bool GetSciteDefaultHome(char *path, unsigned int lenPath) = 0;
 	virtual bool GetSciteUserHome(char *path, unsigned int lenPath) = 0;
-	virtual bool GetDefaultPropertiesFileName(char *pathDefaultProps,
-	        char *pathDefaultDir, unsigned int lenPath) = 0;
-	virtual bool GetUserPropertiesFileName(char *pathUserProps,
-	                                       char *pathUserDir, unsigned int lenPath) = 0;
-	virtual bool GetAbbrevPropertiesFileName(char *pathAbbrevProps,
-	        char *pathDefaultDir, unsigned int lenPath) = 0;
+	bool GetDefaultPropertiesFileName(char *pathDefaultProps,
+	        char *pathDefaultDir, unsigned int lenPath);
+	bool GetUserPropertiesFileName(char *pathUserProps,
+		char *pathUserDir, unsigned int lenPath);
+	bool GetAbbrevPropertiesFileName(char *pathAbbrevProps,
+	        char *pathDefaultDir, unsigned int lenPath);
 	void OpenProperties(int propsFile);
+	int GetMenuCommandAsInt(SString commandName);
 	virtual void Print(bool) {};
 	virtual void PrintSetup() {};
 	CharacterRange GetSelection();
 	void SetSelection(int anchor, int currentPos);
 	//	void SelectionExtend(char *sel, int len, char *notselchar);
 	void GetCTag(char *sel, int len);
-	void RangeExtendAndGrab(Window &wCurrent, char *sel, int len,
-	    int selStart, int selEnd, int lengthDoc, bool (*ischarforsel)(char ch));
-	void SelectionExtend(char *sel, int len, bool (*ischarforsel)(char ch));
-	SString SelectionWord();
+	SString RangeExtendAndGrab(Window &wCurrent, int &selStart, int &selEnd,
+		bool (*ischarforsel)(char ch), bool stripEol = true);
+	SString SelectionExtend(bool (*ischarforsel)(char ch), bool stripEol = true);
+	void FindWordAtCaret(int &start, int &end);
+	bool SelectWordAtCaret();
+	SString SelectionWord(bool stripEol = true);
 	SString SelectionFilename();
 	void SelectionIntoProperties();
-	void SelectionIntoFind();
+	void SelectionIntoFind(bool stripEol = true);
 	virtual void Find() = 0;
 	virtual int WindowMessageBox(Window &w, const SString &m, int style)=0;
 	void FindMessageBox(const SString &msg);
-	void FindNext(bool reverseDirection, bool showWarnings = true);
+	int FindNext(bool reverseDirection, bool showWarnings = true);
+	virtual void FindIncrement() = 0;
+	int IncrementSearchMode();
 	virtual void FindInFiles() = 0;
 	virtual void Replace() = 0;
 	void ReplaceOnce();
-	void ReplaceAll(bool inSelection);
+	int DoReplaceAll(bool inSelection); // returns number of replacements or negative value if error
+	int ReplaceAll(bool inSelection);
+	int ReplaceInBuffers();
 	virtual void DestroyFindReplace() = 0;
 	virtual void GoLineDialog() = 0;
+	virtual bool AbbrevDialog() = 0;
 	virtual void TabSizeDialog() = 0;
 	virtual void ParamGrab() = 0;
 	virtual bool ParametersDialog(bool modal) = 0;
+	bool HandleXml(char ch);
+	SString FindOpenXmlTag(const char sel[], int nSize);
 	void GoMatchingBrace(bool select);
 	void GoMatchingPreprocCond(int direction, bool select);
 	virtual void FindReplace(bool replace) = 0;
-	void OutputAppendString(const char *s, int len = -1);
-	void OutputAppendStringSynchronised(const char *s, int len = -1);
+	void OutputAppendString(const char *s, int len = -1, bool startLine = false);
+	void OutputAppendStringSynchronised(const char *s, int len = -1, bool startLine = false);
 	void MakeOutputVisible();
 	void ClearJobQueue();
 	virtual void Execute();
 	virtual void StopExecute() = 0;
 	void GoMessage(int dir);
 	virtual bool StartCallTip();
+	virtual void FillFunctionDefinition(int pos = -1);
 	void ContinueCallTip();
+	virtual void EliminateDuplicateWords(char *words);
 	virtual bool StartAutoComplete();
 	virtual bool StartAutoCompleteWord(bool onlyOneWord);
 	virtual bool StartExpandAbbreviation();
+	virtual bool StartInsertAbbreviation();
 	virtual bool StartBlockComment();
 	virtual bool StartBoxComment();
 	virtual bool StartStreamComment();
@@ -564,17 +771,23 @@ protected:
 	void MaintainIndentation(char ch);
 	void AutomaticIndentation(char ch);
 	void CharAdded(char ch);
-	void SetTextProperties(PropSet &ps);
+	void SetTextProperties(PropSetFile &ps);
 	void SetFileProperties(PropSet &ps);
 	virtual void UpdateStatusBar(bool bUpdateSlowData);
+	int GetLineLength(int line);
 	int GetCurrentLineNumber();
 	int GetCurrentScrollPosition();
-	virtual void AddCommand(const SString &cmd, const SString &dir, JobSubsystem jobType, bool forceQueue = false);
+	virtual void AddCommand(const SString &cmd, const SString &dir,
+		JobSubsystem jobType, const SString &input = "",
+		int flags = 0);
 	virtual void AboutDialog() = 0;
 	virtual void QuitProgram() = 0;
-	void CloseAllBuffers();
+	void CloseAllBuffers(bool loadingSession = false);
+	int SaveAllBuffers(bool forceQuestion, bool alwaysYes=false);
+	void SaveTitledBuffers();
 	virtual void CopyAsRTF() {};
-	void MenuCommand(int cmdID);
+	void SetLineNumberWidth();
+	void MenuCommand(int cmdID, int source=0);
 	void FoldChanged(int line, int levelNow, int levelPrev);
 	void FoldChanged(int position);
 	void Expand(int &line, bool doExpand, bool force = false,
@@ -590,9 +803,12 @@ protected:
 	virtual void ShowTabBar() = 0;
 	virtual void ShowStatusBar() = 0;
 
-	void BookmarkToggle( int lineno = -1 );
-	void BookmarkNext();
-	void BookmarkPrev();
+	int MarkAll();
+	void BookmarkAdd(int lineno = -1);
+	void BookmarkDelete(int lineno = -1);
+	bool BookmarkPresent(int lineno = -1);
+	void BookmarkToggle(int lineno = -1);
+	void BookmarkNext(bool forwardScan = true);
 	void ToggleOutputVisible();
 	virtual void SizeContentWindows() = 0;
 	virtual void SizeSubWindows() = 0;
@@ -602,6 +818,7 @@ protected:
 	virtual void DestroyMenuItem(int menuNumber, int itemID) = 0;
 	virtual void CheckAMenuItem(int wIDCheckItem, bool val) = 0;
 	virtual void EnableAMenuItem(int wIDCheckItem, bool val) = 0;
+	virtual void CheckMenusClipboard();
 	virtual void CheckMenus();
 	virtual void AddToPopUp(const char *label, int cmd=0, bool enabled=true)=0;
 	void ContextMenu(Window wSource, Point pt, Window wCmd);
@@ -609,7 +826,7 @@ protected:
 	void DeleteFileStackMenu();
 	void SetFileStackMenu();
 	void DropFileStackTop();
-	void AddFileToBuffer(const char *file /*TODO:, CharacterRange selection, int scrollPos */);
+	void AddFileToBuffer(const char *file, int pos);
 	void AddFileToStack(const char *file, CharacterRange selection, int scrollPos);
 	void RemoveFileFromStack(const char *file);
 	RecentFile GetFilePosition();
@@ -634,8 +851,9 @@ protected:
 	void SetLanguageMenu();
 	void SetPropertiesInitial();
 	SString LocaliseString(const char *s, bool retainIfNotFound=true);
-	SString LocaliseMessage(const char *s, const char *param0=0, const char *param1=0);
-	void ReadLocalisation();
+	SString LocaliseMessage(const char *s, const char *param0=0,
+		const char *param1=0, const char *param2=0);
+	virtual void ReadLocalisation();
 	virtual void ReadPropertiesInitial();
 	void SetMonoFont();
 	void SetOverrideLanguage(int cmdID);
@@ -667,10 +885,12 @@ protected:
 	void AskMacroList();
 	bool StartMacroList(const char *words);
 	void ContinueMacroList(const char *stxt);
+	void LoadMRUAndSession(bool allowLoadSession);
 	bool ProcessCommandLine(SString &args, int phase);
 	void EnumProperties(const char *action);
 	void SendOneProperty(const char *kind, const char *key, const char *val);
 	void PropertyFromDirector(const char *arg);
+	void PropertyToDirector(const char *arg);
 
 	// ExtensionAPI
 	sptr_t Send(Pane p, unsigned int msg, uptr_t wParam = 0, sptr_t lParam = 0);
@@ -691,6 +911,11 @@ public:
 
 	void ProcessExecute();
 	WindowID GetID() { return wSciTE.GetID(); }
+
+private:
+	// un-implemented copy-constructor and assignment operator
+	SciTEBase(const SciTEBase&);
+	void operator=(const SciTEBase&);
 };
 
 /// Base size of file I/O operations.
@@ -702,6 +927,7 @@ const int blockSize = 131072;
 #define MB_YESNO	(0x4L)
 #define MB_YESNOCANCEL	(0x3L)
 #define MB_ICONWARNING	(0x30L)
+#define MB_ICONQUESTION (0x20L)
 #define IDOK	(1)
 #define IDCANCEL	(2)
 #define IDYES	(6)
@@ -709,12 +935,13 @@ const int blockSize = 131072;
 #endif
 
 int ControlIDOfCommand(unsigned long);
+bool BuildPath(char *path, const char *dir, const char *fileName,
+	unsigned int lenPath);
 time_t GetModTime(const char *fullPath);
 bool IsUntitledFileName(const char *name);
 void LowerCaseString(char *s);
 void ChopTerminalSlash(char *path);
 int IntFromHexDigit(const char ch);
-ColourDesired ColourFromString(const char *val);
 long ColourOfProperty(PropSet &props, const char *key, ColourDesired colourDefault);
 char *Slash(const char *s);
 unsigned int UnSlash(char *s);
