@@ -2,7 +2,7 @@
 /** @file SciTEIO.cxx
  ** Manage input and output with the system.
  **/
-// Copyright 1998-2002 by Neil Hodgson <neilh@scintilla.org>
+// Copyright 1998-2003 by Neil Hodgson <neilh@scintilla.org>
 // The License.txt file describes the conditions under which this software may be distributed.
 
 #include <stdlib.h>
@@ -31,7 +31,15 @@
 #if PLAT_WIN
 
 #define _WIN32_WINNT  0x0400
+#ifdef _MSC_VER
+// windows.h, et al, use a lot of nameless struct/unions - can't fix it, so allow it
+#pragma warning(disable: 4201)
+#endif
 #include <windows.h>
+#ifdef _MSC_VER
+// okay, that's done, don't allow it in our code
+#pragma warning(default: 4201)
+#endif
 #include <commctrl.h>
 
 // For chdir
@@ -50,32 +58,36 @@
 #include "WindowAccessor.h"
 #include "Scintilla.h"
 #include "Extender.h"
+#include "Utf8_16.h"
 #include "SciTEBase.h"
 
 #ifdef unix
 const char pathSepString[] = "/";
 const char pathSepChar = '/';
+const char listSepString[] = ":";
 const char configFileVisibilityString[] = ".";
 const char propUserFileName[] = ".SciTEUser.properties";
-const char fileRead[]="rb";
-const char fileWrite[]="wb";
+const char fileRead[] = "rb";
+const char fileWrite[] = "wb";
 #endif
 #ifdef __vms
 const char pathSepString[] = "/";
 const char pathSepChar = '/';
+const char listSepString[] = ":";
 const char configFileVisibilityString[] = "";
 const char propUserFileName[] = "SciTEUser.properties";
-const char fileRead[]="r";
-const char fileWrite[]="w";
+const char fileRead[] = "r";
+const char fileWrite[] = "w";
 #endif
 #ifdef WIN32
 // Windows
 const char pathSepString[] = "\\";
 const char pathSepChar = '\\';
+const char listSepString[] = ";";
 const char configFileVisibilityString[] = "";
 const char propUserFileName[] = "SciTEUser.properties";
-const char fileRead[]="rb";
-const char fileWrite[]="wb";
+const char fileRead[] = "rb";
+const char fileWrite[] = "wb";
 #endif
 const char propGlobalFileName[] = "SciTEGlobal.properties";
 const char propAbbrevFileName[] = "abbrev.properties";
@@ -83,8 +95,8 @@ const char propAbbrevFileName[] = "abbrev.properties";
 #define PROPERTIES_EXTENSION	".properties"
 
 static bool IsPropertiesFile(char *filename) {
-	int nameLen = strlen(filename);
-	int propLen = strlen(PROPERTIES_EXTENSION);
+	size_t nameLen = strlen(filename);
+	size_t propLen = strlen(PROPERTIES_EXTENSION);
 	if (nameLen <= propLen)
 		return false;
 	if (EqualCaseInsensitive(filename + nameLen - propLen, PROPERTIES_EXTENSION))
@@ -96,17 +108,21 @@ bool SciTEBase::IsAbsolutePath(const char *path) {
 	if (!path || *path == '\0')
 		return false;
 #ifdef unix
+
 	if (path[0] == '/')
 		return true;
 #endif
 #ifdef __vms
+
 	if (path[0] == '/')
 		return true;
 #endif
 #ifdef WIN32
+
 	if (path[0] == '\\' || path[1] == ':')	// UNC path or drive separator
 		return true;
 #endif
+
 	return false;
 }
 
@@ -233,8 +249,6 @@ void SciTEBase::SetFileName(const char *openName, bool fixCase) {
 		fileExt[0] = '\0';
 	}
 
-	chdir(dirName);
-
 	ReadLocalPropFile();
 
 	props.Set("FilePath", fullPath);
@@ -245,7 +259,7 @@ void SciTEBase::SetFileName(const char *openName, bool fixCase) {
 
 	SetWindowName();
 	if (buffers.buffers)
-		buffers.buffers[buffers.current].Set(fullPath);
+		buffers.buffers[buffers.Current()].Set(fullPath);
 	BuffersMenu();
 }
 
@@ -260,6 +274,8 @@ bool SciTEBase::Exists(const char *dir, const char *path, char *testPath) {
 		strcat(copyPath, pathSepString);
 		strcat(copyPath, path);
 	}
+	if (!copyPath[0])
+		return false;
 	FILE *fp = fopen(copyPath, fileRead);
 	if (!fp)
 		return false;
@@ -269,6 +285,18 @@ bool SciTEBase::Exists(const char *dir, const char *path, char *testPath) {
 	return true;
 }
 
+bool BuildPath(char *path, const char *dir, const char *fileName,
+               unsigned int lenPath) {
+	*path = '\0';
+	if ((strlen(dir) + strlen(pathSepString) + strlen(fileName)) < lenPath) {
+		strcpy(path, dir);
+		strcat(path, pathSepString);
+		strcat(path, fileName);
+		return true;
+	}
+	return false;
+}
+
 time_t GetModTime(const char *fullPath) {
 	if (IsUntitledFileName(fullPath))
 		return 0;
@@ -303,36 +331,112 @@ void SciTEBase::CountLineEnds(int &linesCR, int &linesLF, int &linesCRLF) {
 	}
 }
 
-void SciTEBase::OpenFile(bool initialCmdLine) {
-	FILE *fp = fopen(fullPath, fileRead);
-	if (fp || initialCmdLine) {
-		// If initial run and no fp, just open an empty buffer
-		// with the given name
-		if (fp) {
-			fileModTime = GetModTime(fullPath);
-			fileModLastAsk = fileModTime;
-			SendEditor(SCI_CLEARALL);
-			char data[blockSize];
-			int lenFile = fread(data, 1, sizeof(data), fp);
-			while (lenFile > 0) {
-				SendEditorString(SCI_ADDTEXT, lenFile, data);
-				lenFile = fread(data, 1, sizeof(data), fp);
+static bool isEncodingChar(char ch) {
+	return (ch == '_') || (ch == '-') || (ch == '.') ||
+		(ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') ||
+		(ch >= '0' && ch <= '9');
+}
+
+static bool isSpaceChar(char ch) {
+	return (ch == ' ') || (ch == '\t');
+}
+
+static SString ExtractLine(const char *buf, size_t length) {
+	unsigned int endl = 0;
+	if (length > 0) {
+		while ((endl < length) && (buf[endl] != '\r') && (buf[endl] != '\n')) {
+			endl++;
+		}
+		if (((endl+1) < length) && (buf[endl] == '\r') && (buf[endl+1] == '\n')) {
+			endl++;
+		}
+		if (endl < length) {
+			endl++;
+		}
+	}
+	return SString(buf, 0, endl);
+}
+
+static const char codingCookie[] = "coding";
+
+static UniMode CookieValue(const SString &s) {
+	int posCoding = s.search(codingCookie);
+	if (posCoding >= 0) {
+		posCoding += static_cast<int>(strlen(codingCookie));
+		if ((s[posCoding] == ':') || (s[posCoding] == '=')) {
+			posCoding++;
+			while ((posCoding < static_cast<int>(s.length())) &&
+				(isSpaceChar(s[posCoding]))) {
+				posCoding++;
 			}
-			fclose(fp);
-			if (props.GetInt("eol.auto")) {
-				int linesCR;
-				int linesLF;
-				int linesCRLF;
-				CountLineEnds(linesCR, linesLF, linesCRLF);
-				if ((linesLF > linesCR) && (linesLF > linesCRLF))
-					SendEditor(SCI_SETEOLMODE, SC_EOL_LF);
-				if ((linesCR > linesLF) && (linesCR > linesCRLF))
-					SendEditor(SCI_SETEOLMODE, SC_EOL_CR);
-				if ((linesCRLF > linesLF) && (linesCRLF > linesCR))
-					SendEditor(SCI_SETEOLMODE, SC_EOL_CRLF);
+			size_t endCoding = static_cast<size_t>(posCoding);
+			while ((endCoding < s.length()) &&
+				(isEncodingChar(s[endCoding]))) {
+				endCoding++;
+			}
+			SString code(s.c_str(), posCoding, endCoding);
+			code.lowercase();
+			if (code == "utf-8") {
+				return uniCookie;
 			}
 		}
-	} else {
+	}
+	return uni8Bit;
+}
+
+void SciTEBase::OpenFile(bool suppressMessage) {
+	Utf8_16_Read convert;
+
+	FILE *fp = fopen(fullPath, fileRead);
+	if (fp) {
+		fseek(fp, 0, SEEK_END);
+		int allocation = ftell(fp);
+		SendEditor(SCI_ALLOCATE, allocation + 1000);
+		fseek(fp, 0, SEEK_SET);
+		fileModTime = GetModTime(fullPath);
+		fileModLastAsk = fileModTime;
+		SendEditor(SCI_CLEARALL);
+		char data[blockSize];
+		size_t lenFile = fread(data, 1, sizeof(data), fp);
+		SString l1 = ExtractLine(data, lenFile);
+		SString l2 = ExtractLine(data+l1.length(), lenFile-l1.length());
+		while (lenFile > 0) {
+			lenFile = convert.convert(data, lenFile);
+			SendEditorString(SCI_ADDTEXT, lenFile, convert.getNewBuf());
+			lenFile = fread(data, 1, sizeof(data), fp);
+		}
+		fclose(fp);
+		unicodeMode = static_cast<UniMode>(
+			static_cast<int>(convert.getEncoding()));
+		// Check the first two lines for coding cookies
+		if (unicodeMode == uni8Bit) {
+			unicodeMode = CookieValue(l1);
+			if (unicodeMode == uni8Bit) {
+				unicodeMode = CookieValue(l2);
+			}
+		}
+		if (unicodeMode != uni8Bit) {
+			// Override the code page if Unicode
+			codePage = SC_CP_UTF8;
+		} else {
+			codePage = props.GetInt("code.page");
+		}
+		SendEditor(SCI_SETCODEPAGE, codePage);
+
+		if (props.GetInt("eol.auto")) {
+			int linesCR;
+			int linesLF;
+			int linesCRLF;
+			SetEol();
+			CountLineEnds(linesCR, linesLF, linesCRLF);
+			if (((linesLF >= linesCR) && (linesLF > linesCRLF)) || ((linesLF > linesCR) && (linesLF >= linesCRLF)))
+				SendEditor(SCI_SETEOLMODE, SC_EOL_LF);
+			else if (((linesCR >= linesLF) && (linesCR > linesCRLF)) || ((linesCR > linesLF) && (linesCR >= linesCRLF)))
+				SendEditor(SCI_SETEOLMODE, SC_EOL_CR);
+			else if (((linesCRLF >= linesLF) && (linesCRLF > linesCR)) || ((linesCRLF > linesLF) && (linesCRLF >= linesCR)))
+				SendEditor(SCI_SETEOLMODE, SC_EOL_CRLF);
+		}
+	} else if (!suppressMessage) {
 		SString msg = LocaliseMessage("Could not open file '^0'.", fullPath);
 		WindowMessageBox(wSciTE, msg, MB_OK | MB_ICONWARNING);
 	}
@@ -349,8 +453,11 @@ void SciTEBase::OpenFile(bool initialCmdLine) {
 	Redraw();
 }
 
-bool SciTEBase::Open(const char *file, bool initialCmdLine, 
-	bool forceLoad, bool maySaveIfDirty) {
+bool SciTEBase::PreOpenCheck(const char *) {
+	return false;
+}
+
+bool SciTEBase::Open(const char *file, OpenFlags of) {
 	InitialiseBuffers();
 
 	if (!file) {
@@ -358,21 +465,23 @@ bool SciTEBase::Open(const char *file, bool initialCmdLine,
 		WindowMessageBox(wSciTE, msg, MB_OK | MB_ICONWARNING);
 	}
 #ifdef __vms
-	static char fixedFileName [MAX_PATH];
+	static char fixedFileName[MAX_PATH];
 	strcpy(fixedFileName, VMSToUnixStyle(file));
 	file = fixedFileName;
 #endif
 
-	int index = buffers.GetDocumentByName(file);
+	char absPath[MAX_PATH];
+	AbsolutePath(absPath, file, MAX_PATH);
+	int index = buffers.GetDocumentByName(absPath);
 	if (index >= 0) {
 		SetDocumentAt(index);
 		DeleteFileStackMenu();
 		SetFileStackMenu();
-		if (!forceLoad) // Just rotate into view
+		if (!(of & ofForceLoad)) // Just rotate into view
 			return true;
 	}
 	// See if we can have a buffer for the file to open
-	if (!CanMakeRoom(maySaveIfDirty)) {
+	if (!CanMakeRoom(!(of & ofNoSaveIfDirty))) {
 		return false;
 	}
 
@@ -380,37 +489,37 @@ bool SciTEBase::Open(const char *file, bool initialCmdLine,
 		AddFileToStack(fullPath, GetSelection(), GetCurrentScrollPosition());
 		ClearDocument();
 	} else {
-		if (index < 0 || !forceLoad) { // No new buffer, already opened
+		if (index < 0 || !(of & ofForceLoad)) { // No new buffer, already opened
 			New();
 		}
 	}
 
 	//Platform::DebugPrintf("Opening %s\n", file);
-	SetFileName(file);
+	SetFileName(absPath);
 	overrideExtension = "";
 	ReadProperties();
-	if (useMonoFont) {
-		SetMonoFont();
-	}
+	SetIndentSettings();
 	UpdateBuffersCurrent();
-	if (tabVisible)	// Update tab visibility if the hide one mode is on
-		SizeSubWindows();
+	SizeSubWindows();
 
-	if (initialCmdLine) {
-		if (props.GetInt("save.recent", 0))
-			LoadRecentMenu();
-	}
-	if (initialCmdLine && props.GetInt("buffers") && !fileName[0]) {
-		if (props.GetInt("save.session", 0))
-			LoadSession("");
-	}
 	if (fileName[0]) {
+		SendEditor(SCI_SETREADONLY, 0);
 		SendEditor(SCI_CANCEL);
-		SendEditor(SCI_SETUNDOCOLLECTION, 0);
+		if (of & ofPreserveUndo) {
+			SendEditor(SCI_BEGINUNDOACTION);
+		} else {
+			SendEditor(SCI_SETUNDOCOLLECTION, 0);
+		}
 
-		OpenFile(initialCmdLine);
+		OpenFile(of & ofQuiet);
 
-		SendEditor(SCI_EMPTYUNDOBUFFER);
+		if (of & ofPreserveUndo) {
+			SendEditor(SCI_ENDUNDOACTION);
+		} else {
+			SendEditor(SCI_EMPTYUNDOBUFFER);
+		}
+		isReadOnly = props.GetInt("read.only");
+		SendEditor(SCI_SETREADONLY, isReadOnly);
 	}
 	RemoveFileFromStack(fullPath);
 	DeleteFileStackMenu();
@@ -422,40 +531,36 @@ bool SciTEBase::Open(const char *file, bool initialCmdLine,
 	return true;
 }
 
-void SciTEBase::OpenMultiple(const char *files, bool initialCmdLine, bool forceLoad) {
-	if (*files) {
-		while (*files) {
-			Open(files, initialCmdLine, forceLoad);
-			files = files + strlen(files) + 1;
-		}
-	} else {
-		Open("", initialCmdLine, forceLoad);
-	}
-}
-
-void SciTEBase::OpenSelected() {
+// Returns true if editor should get the focus
+bool SciTEBase::OpenSelected() {
 	char selectedFilename[MAX_PATH];
 	char cTag[200];
 	unsigned long lineNumber = 0;
 
 	SString selName = SelectionFilename();
 	strncpy(selectedFilename, selName.c_str(), MAX_PATH);
-	selectedFilename[MAX_PATH-1] = '\0';
+	selectedFilename[MAX_PATH - 1] = '\0';
 	if (selectedFilename[0] == '\0') {
 		WarnUser(warnWrongFile);
-		return;	// No selection
+		return false;	// No selection
 	}
+	SString fileNameForExtension = ExtensionFileName();
+	SString openSuffix = props.GetNewExpand("open.suffix.", fileNameForExtension.c_str());
+	strcat(selectedFilename, openSuffix.c_str());
 
-	if (EqualCaseInsensitive(selectedFilename, fileName)) {
+	if (EqualCaseInsensitive(selectedFilename, fileName) || EqualCaseInsensitive(selectedFilename, fullPath)) {
 		WarnUser(warnWrongFile);
-		return;	// Do not open if it is the current file!
+		return true;	// Do not open if it is the current file!
 	}
 
 	cTag[0] = '\0';
 	if (IsPropertiesFile(fileName) &&
-	        strlen(fileName) + strlen(PROPERTIES_EXTENSION) < MAX_PATH) {
-		// We are in a properties file, we append the correct extension
-		// to open the include
+			strchr(selectedFilename, '.') == 0 &&
+	        strlen(selectedFilename) + strlen(PROPERTIES_EXTENSION) < MAX_PATH) {
+		// We are in a properties file and try to open a file without extension,
+		// we suppose we want to open an imported .properties file
+		// So we append the correct extension to open the included file.
+		// Maybe we should check if the filename is preceded by "import"...
 		strcat(selectedFilename, PROPERTIES_EXTENSION);
 	} else {
 		// Check if we have a line number (error message or grep result)
@@ -466,7 +571,7 @@ void SciTEBase::OpenSelected() {
 		if (endPath) {	// Visual Studio error message: F:\scite\src\SciTEBase.h(312):	bool Exists(
 			lineNumber = atol(endPath + 1);
 		} else {
-			char *endPath = strchr(selectedFilename + 2, ':');	// Skip Windows' drive separator
+			endPath = strchr(selectedFilename + 2, ':');	// Skip Windows' drive separator
 			if (endPath) {	// grep -n line, perhaps gcc too: F:\scite\src\SciTEBase.h:312:	bool Exists(
 				lineNumber = atol(endPath + 1);
 			}
@@ -483,14 +588,15 @@ void SciTEBase::OpenSelected() {
 		        strncmp(selectedFilename, "news:", 5) == 0 ||
 		        strncmp(selectedFilename, "mailto:", 7) == 0) {
 			SString cmd = selectedFilename;
-			AddCommand(cmd, "", jobShell, false);
-			return;	// Job is done
+			AddCommand(cmd, "", jobShell);
+			return false;	// Job is done
 		}
 #endif
 
 		// Support the ctags format
+
 		if (lineNumber == 0) {
-			GetCTag(cTag, 200);
+			GetCTag(cTag, sizeof(cTag));
 		}
 	}
 
@@ -500,19 +606,45 @@ void SciTEBase::OpenSelected() {
 	// filename is an absolute pathname
 	if (!IsAbsolutePath(selectedFilename)) {
 		GetDocumentDirectory(path, sizeof(path));
+		// If not there, look in openpath
+		if (!Exists(path, selectedFilename, NULL)) {
+			SString openPath = props.GetNewExpand(
+				"openpath.", fileNameForExtension.c_str());
+			while (openPath.length()) {
+				SString tryPath(openPath);
+				int sepIndex = tryPath.search(listSepString);
+				if (sepIndex > 0) {
+					tryPath.remove(sepIndex, 0);
+					openPath.remove(0,sepIndex+1);
+				} else {
+					openPath.clear();
+				}
+				if (Exists(tryPath.c_str(), selectedFilename, NULL)) {
+					strncpy(path, tryPath.c_str(), sizeof(path)-1);
+					path[sizeof(path)-1] = '\0';
+					break;
+				}
+			}
+		}
 	}
 	if (Exists(path, selectedFilename, path)) {
-		if (Open(path, false)) {
+		if (Open(path)) {
 			if (lineNumber > 0) {
 				SendEditor(SCI_GOTOLINE, lineNumber - 1);
 			} else if (cTag[0] != '\0') {
-				findWhat = cTag;
-				FindNext(false);
+				if (atol(cTag) > 0) {
+					SendEditor(SCI_GOTOLINE, atol(cTag) - 1);
+				} else {
+					findWhat = cTag;
+					FindNext(false);
+				}
 			}
+			return true;
 		}
 	} else {
 		WarnUser(warnWrongFile);
 	}
+	return false;
 }
 
 void SciTEBase::Revert() {
@@ -530,25 +662,28 @@ void SciTEBase::CheckReload() {
 		//Platform::DebugPrintf("Times are %d %d\n", fileModTime, newModTime);
 		if (newModTime > fileModTime) {
 			RecentFile rf = GetFilePosition();
-			if (isDirty) {
-				static bool entered = false; // Stop reentrancy
-				if (!entered && (0 == dialogsOnScreen) && (newModTime != fileModLastAsk)) {
-					entered = true;
-					SString msg = LocaliseMessage(
-						"The file '^0' has been modified. Should it be reloaded?", 
-						fullPathToCheck);
-					dialogsOnScreen++;
+			OpenFlags of = props.GetInt("reload.preserves.undo") ? ofPreserveUndo : ofNone;
+			if (isDirty || props.GetInt("are.you.sure.on.reload") != 0) {
+				if ((0 == dialogsOnScreen) && (newModTime != fileModLastAsk)) {
+					SString msg;
+					if (isDirty) {
+						msg = LocaliseMessage(
+							  "The file '^0' has been modified. Should it be reloaded?",
+							  fullPathToCheck);
+					} else {
+						msg = LocaliseMessage(
+							  "The file '^0' has been modified outside SciTE. Should it be reloaded?",
+							  fileName);
+					}
 					int decision = WindowMessageBox(wSciTE, msg, MB_YESNO);
-					dialogsOnScreen--;
 					if (decision == IDYES) {
-						Open(fullPathToCheck, false, true, false);
+						Open(fullPathToCheck, static_cast<OpenFlags>(of | ofForceLoad));
 						DisplayAround(rf);
 					}
 					fileModLastAsk = newModTime;
-					entered = false;
 				}
 			} else {
-				Open(fullPathToCheck, false, true);
+				Open(fullPathToCheck, static_cast<OpenFlags>(of | ofForceLoad));
 				DisplayAround(rf);
 			}
 		}
@@ -559,17 +694,14 @@ void SciTEBase::Activate(bool activeApp) {
 	if (activeApp) {
 		CheckReload();
 	} else {
-		if (isDirty) {
-			if (props.GetInt("save.on.deactivate") && fileName[0]) {
-				// Trying to save at deactivation when no file name -> dialogs
-				Save();
-			}
+		if (props.GetInt("save.on.deactivate")) {
+			SaveTitledBuffers();
 		}
 	}
 }
 
 int SciTEBase::SaveIfUnsure(bool forceQuestion) {
-	if (isDirty) {
+	if ((isDirty) && (LengthDocument() || fileName[0] || forceQuestion)) {
 		if (props.GetInt("are.you.sure", 1) ||
 		        IsUntitledFileName(fullPath) ||
 		        forceQuestion) {
@@ -579,9 +711,7 @@ int SciTEBase::SaveIfUnsure(bool forceQuestion) {
 			} else {
 				msg = LocaliseMessage("Save changes to (Untitled)?");
 			}
-			dialogsOnScreen++;
-			int decision = WindowMessageBox(wSciTE, msg, MB_YESNOCANCEL);
-			dialogsOnScreen--;
+			int decision = WindowMessageBox(wSciTE, msg, MB_YESNOCANCEL | MB_ICONQUESTION);
 			if (decision == IDYES) {
 				if (!Save())
 					decision = IDCANCEL;
@@ -596,16 +726,17 @@ int SciTEBase::SaveIfUnsure(bool forceQuestion) {
 }
 
 int SciTEBase::SaveIfUnsureAll(bool forceQuestion) {
-	UpdateBuffersCurrent();	// Ensure isDirty copied
-	for (int i = 0; i < buffers.length; i++) {
-		if (buffers.buffers[i].isDirty) {
-			SetDocumentAt(i);
-			if (SaveIfUnsure(forceQuestion) == IDCANCEL)
-				return IDCANCEL;
-		}
+	if (SaveAllBuffers(forceQuestion) == IDCANCEL) {
+		return IDCANCEL;
 	}
-	if (props.GetInt("save.recent", 0))
+	if (props.GetInt("save.recent", 0)) {
+		for (int i = 0; i < buffers.length; ++i) {
+			Buffer buff = buffers.buffers[i];
+			AddFileToStack(buff.FullPath(),
+			               buff.selection, buff.scrollPosition);
+		}
 		SaveRecentStack();
+	}
 	if (props.GetInt("buffers") && props.GetInt("save.session", 0))
 		SaveSession("");
 
@@ -631,60 +762,105 @@ int SciTEBase::SaveIfUnsureForBuilt() {
 	return IDYES;
 }
 
-int StripTrailingSpaces(char *data, int ds, bool lastBlock) {
-	int lastRead = 0;
-	char *w = data;
-
-	for (int i = 0; i < ds; i++) {
-		char ch = data[i];
-		if ((ch == ' ') || (ch == '\t')) {
-			// Swallow those spaces
+void SciTEBase::StripTrailingSpaces() {
+	int maxLines = SendEditor(SCI_GETLINECOUNT);
+	for (int line = 0; line < maxLines; line++) {
+		int lineStart = SendEditor(SCI_POSITIONFROMLINE, line);
+		int lineEnd = SendEditor(SCI_GETLINEENDPOSITION, line);
+		int i = lineEnd-1;
+		char ch = static_cast<char>(SendEditor(SCI_GETCHARAT, i));
+		while ((i >= lineStart) && ((ch == ' ') || (ch == '\t'))) {
+			i--;
+			ch = static_cast<char>(SendEditor(SCI_GETCHARAT, i));
 		}
-		else if ((ch == '\r') || (ch == '\n')) {
-			*w++ = ch;
-			lastRead = i + 1;
-		} else {
-			while (lastRead < i) {
-				*w++ = data[lastRead++];
-			}
-			*w++ = ch;
-			lastRead = i + 1;
+		if (i < (lineEnd-1)) {
+			SendEditor(SCI_SETTARGETSTART, i + 1);
+			SendEditor(SCI_SETTARGETEND, lineEnd);
+			SendEditorString(SCI_REPLACETARGET, 0, "");
 		}
 	}
-	// If a non-final block, then preserve whitespace at end of block as it may be significant.
-	if (!lastBlock) {
-		while (lastRead < ds) {
-			*w++ = data[lastRead++];
-		}
-	}
-	return w - data;
 }
 
-// Returns false only if cancelled
+void SciTEBase::EnsureFinalNewLine() {
+	int maxLines = SendEditor(SCI_GETLINECOUNT);
+	bool appendNewLine = maxLines == 1;
+	int endDocument = SendEditor(SCI_POSITIONFROMLINE, maxLines);
+	if (maxLines > 1) {
+		appendNewLine = endDocument > SendEditor(SCI_POSITIONFROMLINE, maxLines-1);
+	}
+	if (appendNewLine) {
+		const char *eol = "\n";
+		switch (SendEditor(SCI_GETEOLMODE)) {
+			case SC_EOL_CRLF:
+				eol = "\r\n";
+				break;
+			case SC_EOL_CR:
+				eol = "\r";
+				break;
+		}
+		SendEditorString(SCI_INSERTTEXT, endDocument, eol);
+	}
+}
+
+/**
+ * Writes the buffer to the given filename.
+ */
+bool SciTEBase::SaveBuffer(const char *saveName) {
+	bool retVal = false;
+	// Perform clean ups on text before saving
+	SendEditor(SCI_BEGINUNDOACTION);
+	if (props.GetInt("strip.trailing.spaces"))
+		StripTrailingSpaces();
+	if (props.GetInt("ensure.final.line.end"))
+		EnsureFinalNewLine();
+	if (props.GetInt("ensure.consistent.line.ends"))
+		SendEditor(SCI_CONVERTEOLS, SendEditor(SCI_GETEOLMODE));
+
+	if (extender)
+		extender->OnBeforeSave(saveName);
+
+	SendEditor(SCI_ENDUNDOACTION);
+
+	Utf8_16_Write convert;
+	if (unicodeMode != uniCookie) {	// Save file with cookie without BOM.
+		convert.setEncoding(static_cast<Utf8_16::encodingType>(
+			static_cast<int>(unicodeMode)));
+	}
+
+	FILE *fp = convert.fopen(saveName, fileWrite);
+	if (fp) {
+		char data[blockSize + 1];
+		int lengthDoc = LengthDocument();
+		retVal = true;
+		for (int i = 0; i < lengthDoc; i += blockSize) {
+			int grabSize = lengthDoc - i;
+			if (grabSize > blockSize)
+				grabSize = blockSize;
+			// TODO: This is wrong as it can retrieve partial UTF-8 characters.
+			GetRange(wEditor, i, i + grabSize, data);
+			size_t written = convert.fwrite(data, grabSize);
+			if (written == 0) {
+				retVal = false;
+				break;
+			}
+		}
+		convert.fclose();
+
+		if (extender)
+			extender->OnSave(saveName);
+	}
+	UpdateStatusBar(true);
+	return retVal;
+}
+
+// Returns false if cancelled or failed to save
 bool SciTEBase::Save() {
 	if (fileName[0]) {
 		if (props.GetInt("save.deletes.first")) {
 			unlink(fullPath);
 		}
 
-		FILE *fp = fopen(fullPath, fileWrite);
-		if (fp) {
-			char data[blockSize + 1];
-			int lengthDoc = LengthDocument();
-			for (int i = 0; i < lengthDoc; i += blockSize) {
-				int grabSize = lengthDoc - i;
-				if (grabSize > blockSize)
-					grabSize = blockSize;
-				GetRange(wEditor, i, i + grabSize, data);
-				if (props.GetInt("strip.trailing.spaces"))
-					grabSize = StripTrailingSpaces(
-					               data, grabSize, grabSize != blockSize);
-				fwrite(data, grabSize, 1, fp);
-			}
-			fclose(fp);
-
-			//MoveFile(fullPath, fullPath);
-
+		if (SaveBuffer(fullPath)) {
 			fileModTime = GetModTime(fullPath);
 			SendEditor(SCI_SETSAVEPOINT);
 			if (IsPropertiesFile(fileName)) {
@@ -693,16 +869,18 @@ bool SciTEBase::Save() {
 				ReadLocalPropFile();
 				ReadAbbrevPropFile();
 				ReadProperties();
-				useMonoFont = false;
 				SetWindowName();
 				BuffersMenu();
 				Redraw();
 			}
 		} else {
-			SString msg = LocaliseMessage("Could not save file '^0'.", fullPath);
-			dialogsOnScreen++;
-			WindowMessageBox(wSciTE, msg, MB_OK | MB_ICONWARNING);
-			dialogsOnScreen--;
+			SString msg = LocaliseMessage(
+				"Could not save file '^0'. Save under a different name?", fullPath);
+			int decision = WindowMessageBox(wSciTE, msg, MB_YESNO | MB_ICONWARNING);
+			if (decision == IDYES) {
+				return SaveAs();
+			}
+			return false;
 		}
 		return true;
 	} else {
@@ -710,12 +888,11 @@ bool SciTEBase::Save() {
 	}
 }
 
-bool SciTEBase::SaveAs(char *file) {
+bool SciTEBase::SaveAs(const char *file) {
 	if (file && *file) {
 		SetFileName(file);
 		Save();
 		ReadProperties();
-		useMonoFont = false;
 		SendEditor(SCI_CLEARDOCUMENTSTYLE);
 		SendEditor(SCI_COLOURISE, 0, -1);
 		Redraw();
