@@ -2,7 +2,7 @@
 /** @file SciTEProps.cxx
  ** Properties management.
  **/
-// Copyright 1998-2002 by Neil Hodgson <neilh@scintilla.org>
+// Copyright 1998-2004 by Neil Hodgson <neilh@scintilla.org>
 // The License.txt file describes the conditions under which this software may be distributed.
 
 #include <stdlib.h>
@@ -11,6 +11,7 @@
 #include <stdio.h>
 #include <fcntl.h>
 #include <time.h>
+#include <locale.h>
 
 #include "Platform.h"
 
@@ -34,7 +35,15 @@ const char menuAccessIndicator[] = "_";
 #if PLAT_WIN
 
 #define _WIN32_WINNT  0x0400
+#ifdef _MSC_VER
+// windows.h, et al, use a lot of nameless struct/unions - can't fix it, so allow it
+#pragma warning(disable: 4201)
+#endif
 #include <windows.h>
+#ifdef _MSC_VER
+// okay, that's done, don't allow it in our code
+#pragma warning(default: 4201)
+#endif
 #include <commctrl.h>
 
 // For getcwd
@@ -44,6 +53,9 @@ const char menuAccessIndicator[] = "_";
 #ifdef __BORLANDC__
 #include <dir.h>
 #endif
+#ifdef __DMC__
+#include <dir.h>
+#endif
 
 const char menuAccessIndicator[] = "&";
 
@@ -57,7 +69,7 @@ const char menuAccessIndicator[] = "&";
 #include "Extender.h"
 #include "SciTEBase.h"
 
-PropSetFile::PropSetFile() {}
+PropSetFile::PropSetFile(bool lowerKeys_) : lowerKeys(lowerKeys_) {}
 
 PropSetFile::~PropSetFile() {}
 
@@ -93,10 +105,19 @@ static bool GetFullLine(const char *&fpc, int &lenData, char *s, int len) {
 	return false;
 }
 
+static bool IsSpaceOrTab(char ch) {
+	return (ch == ' ') || (ch == '\t');
+}
+
+static bool IsCommentLine(const char *line) {
+	while (IsSpaceOrTab(*line)) ++line;
+	return (*line == '#');
+}
+
 bool PropSetFile::ReadLine(const char *lineBuffer, bool ifIsTrue, const char *directoryForImports,
                            SString imports[], int sizeImports) {
 	//UnSlash(lineBuffer);
-	if (isalnum(lineBuffer[0]))    // If clause ends with first non-indented line
+	if (!IsSpaceOrTab(lineBuffer[0]))    // If clause ends with first non-indented line
 		ifIsTrue = true;
 	if (isprefix(lineBuffer, "if ")) {
 		const char *expr = lineBuffer + strlen("if") + 1;
@@ -116,9 +137,7 @@ bool PropSetFile::ReadLine(const char *lineBuffer, bool ifIsTrue, const char *di
 				}
 			}
 		}
-	} else if (IsAlphabetic(lineBuffer[0]) || (lineBuffer[0] >= '0' && lineBuffer[0] <= '9')) {
-		Set(lineBuffer);
-	} else if (isspace(lineBuffer[0]) && ifIsTrue) {
+	} else if (ifIsTrue && !IsCommentLine(lineBuffer)) {
 		Set(lineBuffer);
 	}
 	return ifIsTrue;
@@ -131,6 +150,13 @@ void PropSetFile::ReadFromMemory(const char *data, int len, const char *director
 	bool ifIsTrue = true;
 	while (len > 0) {
 		GetFullLine(pd, len, lineBuffer, sizeof(lineBuffer));
+		if (lowerKeys) {
+			for (int i=0; lineBuffer[i] && (lineBuffer[i] != '='); i++) {
+				if ((lineBuffer[i] >= 'A') && (lineBuffer[i] <= 'Z')) {
+					lineBuffer[i] = static_cast<char>(lineBuffer[i] - 'A' + 'a');
+				}
+			}
+		}
 		ifIsTrue = ReadLine(lineBuffer, ifIsTrue, directoryForImports, imports, sizeImports);
 	}
 }
@@ -140,7 +166,7 @@ bool PropSetFile::Read(const char *filename, const char *directoryForImports,
 	FILE *rcfile = fopen(filename, fileRead);
 	if (rcfile) {
 		char propsData[60000];
-		int lenFile = fread(propsData, 1, sizeof(propsData), rcfile);
+		int lenFile = static_cast<int>(fread(propsData, 1, sizeof(propsData), rcfile));
 		fclose(rcfile);
 		ReadFromMemory(propsData, lenFile, directoryForImports, imports, sizeImports);
 		return true;
@@ -148,6 +174,12 @@ bool PropSetFile::Read(const char *filename, const char *directoryForImports,
 	return false;
 }
 
+void PropSetFile::SetInteger(const char *key, sptr_t i) {
+	char tmp[32];
+	sprintf(tmp, "%d", static_cast<int>(i));
+	Set(key, tmp);
+}
+
 void SciTEBase::SetImportMenu() {
 	for (int i = 0; i < importMax; i++) {
 		DestroyMenuItem(menuOptions, importCmdID + i);
@@ -187,16 +219,39 @@ void SciTEBase::SetLanguageMenu() {
 		int itemID = languageCmdID + item;
 		SString entry = LocaliseString(languageMenu[item].menuItem.c_str());
 		if (languageMenu[item].menuKey.length()) {
+#if PLAT_GTK
+			entry += " ";
+#else
 			entry += "\t";
+#endif
 			entry += languageMenu[item].menuKey;
 		}
-		SetMenuItem(menuLanguage, item, itemID, entry.c_str());
+		if (entry[0] != '#') {
+			SetMenuItem(menuLanguage, item, itemID, entry.c_str());
+		}
 	}
 }
 
 const char propFileName[] = "SciTE.properties";
 
 void SciTEBase::ReadGlobalPropFile() {
+#ifdef unix
+	extern char **environ;
+	char **e=environ;
+#else
+	char **e=_environ;
+#endif
+	for (; *e; e++) {
+		char key[1024];
+		char *k=*e;
+		char *v=strchr(k,'=');
+		if (v && (static_cast<size_t>(v-k) < sizeof(key))) {
+			memcpy(key, k, v-k);
+			key[v-k] = '\0';
+			propsEmbed.Set(key, v+1);
+		}
+	}
+
 	for (int stackPos = 0; stackPos < importMax; stackPos++) {
 		importFiles[stackPos] = "";
 	}
@@ -228,7 +283,7 @@ void SciTEBase::ReadAbbrevPropFile() {
 }
 
 void ChopTerminalSlash(char *path) {
-	int endOfPath = strlen(path) - 1;
+	size_t endOfPath = strlen(path) - 1;
 	if (path[endOfPath] == pathSepChar) {
 		path[endOfPath] = '\0';
 	}
@@ -269,27 +324,36 @@ void SciTEBase::ReadLocalPropFile() {
 }
 
 int IntFromHexDigit(const char ch) {
-	if (isdigit(ch))
+	if (isdigit(ch)) {
 		return ch - '0';
-	else if (ch >= 'A' && ch <= 'F')
+	} else if (ch >= 'A' && ch <= 'F') {
 		return ch - 'A' + 10;
-	else if (ch >= 'a' && ch <= 'f')
+	} else if (ch >= 'a' && ch <= 'f') {
 		return ch - 'a' + 10;
-	else
+	} else {
 		return 0;
+	}
 }
 
-ColourDesired ColourFromString(const char *val) {
-	int r = IntFromHexDigit(val[1]) * 16 + IntFromHexDigit(val[2]);
-	int g = IntFromHexDigit(val[3]) * 16 + IntFromHexDigit(val[4]);
-	int b = IntFromHexDigit(val[5]) * 16 + IntFromHexDigit(val[6]);
-	return ColourDesired(r, g, b);
+int IntFromHexByte(const char *hexByte) {
+	return IntFromHexDigit(hexByte[0]) * 16 + IntFromHexDigit(hexByte[1]);
+}
+
+static long ColourFromString(const SString &s) {
+	if (s.length()) {
+		int r = IntFromHexByte(s.c_str() + 1);
+		int g = IntFromHexByte(s.c_str() + 3);
+		int b = IntFromHexByte(s.c_str() + 5);
+		return ColourDesired(r, g, b).AsLong();
+	} else {
+		return ColourDesired().AsLong();
+	}
 }
 
 long ColourOfProperty(PropSet &props, const char *key, ColourDesired colourDefault) {
 	SString colour = props.Get(key);
 	if (colour.length()) {
-		return ColourFromString(colour.c_str()).AsLong();
+		return ColourFromString(colour);
 	}
 	return colourDefault.AsLong();
 }
@@ -300,10 +364,10 @@ long ColourOfProperty(PropSet &props, const char *key, ColourDesired colourDefau
  * @return NULL if the end of the list is met, else, it points to the next item.
  */
 const char *SciTEBase::GetNextPropItem(
-    const char *pStart,    	/**< the property string to parse for the first call,
-        							 * pointer returned by the previous call for the following. */
-    char *pPropItem,    	///< pointer on a buffer receiving the requested prop item
-    int maxLen)			///< size of the above buffer
+	const char *pStart,	/**< the property string to parse for the first call,
+						 * pointer returned by the previous call for the following. */
+	char *pPropItem,	///< pointer on a buffer receiving the requested prop item
+	int maxLen)			///< size of the above buffer
 {
 	int size = maxLen - 1;
 
@@ -325,19 +389,34 @@ const char *SciTEBase::GetNextPropItem(
 }
 
 StyleDefinition::StyleDefinition(const char *definition) :
-		size(0), fore(0), back(ColourDesired(0xff, 0xff, 0xff)), bold(false), italics(false),
-eolfilled(false), underlined(false), caseForce(SC_CASE_MIXED), visible(true) {
-	specified = sdNone;
+		size(0), fore("#000000"), back("#FFFFFF"),
+		bold(false), italics(false), eolfilled(false), underlined(false),
+		caseForce(SC_CASE_MIXED),
+		visible(true), changeable(true),
+		specified(sdNone) {
+	ParseStyleDefinition(definition);
+}
+
+bool StyleDefinition::ParseStyleDefinition(const char *definition) {
+	if (definition == 0 || *definition == '\0') {
+		return false;
+	}
 	char *val = StringDup(definition);
 	//Platform::DebugPrintf("Style %d is [%s]\n", style, val);
 	char *opt = val;
 	while (opt) {
+		// Find attribute separator
 		char *cpComma = strchr(opt, ',');
-		if (cpComma)
+		if (cpComma) {
+			// If found, we terminate the current attribute (opt) string
 			*cpComma = '\0';
+		}
+		// Find attribute name/value separator
 		char *colon = strchr(opt, ':');
-		if (colon)
+		if (colon) {
+			// If found, we terminate the current attribute name and point on the value
 			*colon++ = '\0';
+		}
 		if (0 == strcmp(opt, "italics")) {
 			specified = static_cast<flags>(specified | sdItalics);
 			italics = true;
@@ -357,14 +436,15 @@ eolfilled(false), underlined(false), caseForce(SC_CASE_MIXED), visible(true) {
 		if (0 == strcmp(opt, "font")) {
 			specified = static_cast<flags>(specified | sdFont);
 			font = colon;
+			font.substitute('|', ',');
 		}
 		if (0 == strcmp(opt, "fore")) {
 			specified = static_cast<flags>(specified | sdFore);
-			fore = ColourFromString(colon);
+			fore = colon;
 		}
 		if (0 == strcmp(opt, "back")) {
 			specified = static_cast<flags>(specified | sdBack);
-			back = ColourFromString(colon);
+			back = colon;
 		}
 		if (0 == strcmp(opt, "size")) {
 			specified = static_cast<flags>(specified | sdSize);
@@ -389,10 +469,12 @@ eolfilled(false), underlined(false), caseForce(SC_CASE_MIXED), visible(true) {
 		if (0 == strcmp(opt, "case")) {
 			specified = static_cast<flags>(specified | sdCaseForce);
 			caseForce = SC_CASE_MIXED;
-			if (*colon == 'u')
-				caseForce = SC_CASE_UPPER;
-			else if (*colon == 'l')
-				caseForce = SC_CASE_LOWER;
+			if (colon) {
+				if (*colon == 'u')
+					caseForce = SC_CASE_UPPER;
+				else if (*colon == 'l')
+					caseForce = SC_CASE_LOWER;
+			}
 		}
 		if (0 == strcmp(opt, "visible")) {
 			specified = static_cast<flags>(specified | sdVisible);
@@ -416,6 +498,15 @@ eolfilled(false), underlined(false), caseForce(SC_CASE_MIXED), visible(true) {
 			opt = 0;
 	}
 	delete []val;
+	return true;
+}
+
+long StyleDefinition::ForeAsLong() {
+	return ColourFromString(fore);
+}
+
+long StyleDefinition::BackAsLong() {
+	return ColourFromString(back);
 }
 
 void SciTEBase::SetOneStyle(Window &win, int style, const char *s) {
@@ -425,11 +516,12 @@ void SciTEBase::SetOneStyle(Window &win, int style, const char *s) {
 	if (sd.specified & StyleDefinition::sdBold)
 		Platform::SendScintilla(win.GetID(), SCI_STYLESETBOLD, style, sd.bold ? 1 : 0);
 	if (sd.specified & StyleDefinition::sdFont)
-		Platform::SendScintilla(win.GetID(), SCI_STYLESETFONT, style, reinterpret_cast<long>(sd.font.c_str()));
+		Platform::SendScintillaPointer(win.GetID(), SCI_STYLESETFONT, style,
+			const_cast<char *>(sd.font.c_str()));
 	if (sd.specified & StyleDefinition::sdFore)
-		Platform::SendScintilla(win.GetID(), SCI_STYLESETFORE, style, sd.fore.AsLong());
+		Platform::SendScintilla(win.GetID(), SCI_STYLESETFORE, style, sd.ForeAsLong());
 	if (sd.specified & StyleDefinition::sdBack)
-		Platform::SendScintilla(win.GetID(), SCI_STYLESETBACK, style, sd.back.AsLong());
+		Platform::SendScintilla(win.GetID(), SCI_STYLESETBACK, style, sd.BackAsLong());
 	if (sd.specified & StyleDefinition::sdSize)
 		Platform::SendScintilla(win.GetID(), SCI_STYLESETSIZE, style, sd.size);
 	if (sd.specified & StyleDefinition::sdEOLFilled)
@@ -505,7 +597,7 @@ static int FileLength(const char *path) {
 void SciTEBase::ReadAPI(const SString &fileNameForExtension) {
 	SString apisFileNames = props.GetNewExpand("api.",
 	                        fileNameForExtension.c_str());
-	int nameLength = apisFileNames.length();
+	size_t nameLength = apisFileNames.length();
 	if (nameLength) {
 		apisFileNames.substitute(';', '\0');
 		const char *apiFileName = apisFileNames.c_str();
@@ -554,11 +646,52 @@ SString SciTEBase::FindLanguageProperty(const char *pattern, const char *default
 	return ret;
 }
 
+/**
+ * A list of all the properties that should be forwarded to Scintilla lexers.
+ */
+static const char *propertiesToForward[] = {
+	"fold",
+	"fold.comment",
+	"fold.comment.python",
+	"fold.compact",
+	"fold.at.else",
+	"fold.html",
+	"fold.html.preprocessor",
+	"fold.preprocessor",
+	"fold.quotes.python",
+	"fold.verilog.flags",
+	"styling.within.preprocessor",
+	"tab.timmy.whinge.level",
+	"asp.default.language",
+	"html.tags.case.sensitive",
+	"ps.level",
+	"ps.tokenize",
+
+	"lexer.tex.interface.default",
+	"lexer.tex.comment.process",
+	"lexer.tex.auto.if",
+	"lexer.tex.use.keywords",
+	"lexer.metapost.interface.default",
+	"lexer.metapost.comment.process",
+	"nsis.uservars",
+	"nsis.ignorecase",
+
+	0,
+};
+
 void SciTEBase::ReadProperties() {
 	SString fileNameForExtension = ExtensionFileName();
 
+	SString modulePath = props.GetNewExpand("lexerpath.",
+	    fileNameForExtension.c_str());
+	if (modulePath.length())
+	    SendEditorString(SCI_LOADLEXERLIBRARY, 0, modulePath.c_str());
 	language = props.GetNewExpand("lexer.", fileNameForExtension.c_str());
-	SendEditorString(SCI_SETLEXERLANGUAGE, 0, language.c_str());
+	if (language.length())
+	    SendEditorString(SCI_SETLEXERLANGUAGE, 0, language.c_str());
+	else
+	    SendEditorString(SCI_SETLEXER, 0, SCLEX_CONTAINER);
+
 	lexLanguage = SendEditor(SCI_GETLEXER);
 
 	if ((lexLanguage == SCLEX_HTML) || (lexLanguage == SCLEX_XML) ||
@@ -569,20 +702,16 @@ void SciTEBase::ReadProperties() {
 
 	SendOutput(SCI_SETLEXER, SCLEX_ERRORLIST);
 
-	apis.Clear();
-
 	SString kw0 = props.GetNewExpand("keywords.", fileNameForExtension.c_str());
 	SendEditorString(SCI_SETKEYWORDS, 0, kw0.c_str());
-	SString kw1 = props.GetNewExpand("keywords2.", fileNameForExtension.c_str());
-	SendEditorString(SCI_SETKEYWORDS, 1, kw1.c_str());
-	SString kw2 = props.GetNewExpand("keywords3.", fileNameForExtension.c_str());
-	SendEditorString(SCI_SETKEYWORDS, 2, kw2.c_str());
-	SString kw3 = props.GetNewExpand("keywords4.", fileNameForExtension.c_str());
-	SendEditorString(SCI_SETKEYWORDS, 3, kw3.c_str());
-	SString kw4 = props.GetNewExpand("keywords5.", fileNameForExtension.c_str());
-	SendEditorString(SCI_SETKEYWORDS, 4, kw4.c_str());
-	SString kw5 = props.GetNewExpand("keywords6.", fileNameForExtension.c_str());
-	SendEditorString(SCI_SETKEYWORDS, 5, kw5.c_str());
+
+	for (int wl = 1; wl <= KEYWORDSET_MAX; wl++) {
+		SString kwk(wl+1);
+		kwk += '.';
+		kwk.insert(0, "keywords");
+		SString kw = props.GetNewExpand(kwk.c_str(), fileNameForExtension.c_str());
+		SendEditorString(SCI_SETKEYWORDS, wl, kw.c_str());
+	}
 
 	char homepath[MAX_PATH + 20];
 	if (GetSciteDefaultHome(homepath, sizeof(homepath))) {
@@ -592,34 +721,38 @@ void SciTEBase::ReadProperties() {
 		props.Set("SciteUserHome", homepath);
 	}
 
-	ForwardPropertyToEditor("fold");
-	ForwardPropertyToEditor("fold.compact");
-	ForwardPropertyToEditor("fold.comment");
-	ForwardPropertyToEditor("fold.comment.python");
-	ForwardPropertyToEditor("fold.quotes.python");
-	ForwardPropertyToEditor("fold.html");
-	ForwardPropertyToEditor("styling.within.preprocessor");
-	ForwardPropertyToEditor("tab.timmy.whinge.level");
-	ForwardPropertyToEditor("asp.default.language");
+	for (size_t i=0; propertiesToForward[i]; i++) {
+		ForwardPropertyToEditor(propertiesToForward[i]);
+	}
 
-	ReadAPI(fileNameForExtension);
+	if (apisFileNames != props.GetNewExpand("api.",	fileNameForExtension.c_str())) {
+		apis.Clear();
+		ReadAPI(fileNameForExtension);
+		apisFileNames = props.GetNewExpand("api.", fileNameForExtension.c_str());
+	}
 
 	if (!props.GetInt("eol.auto")) {
-		SString eol_mode = props.Get("eol.mode");
-		if (eol_mode == "LF") {
-			SendEditor(SCI_SETEOLMODE, SC_EOL_LF);
-		} else if (eol_mode == "CR") {
-			SendEditor(SCI_SETEOLMODE, SC_EOL_CR);
-		} else if (eol_mode == "CRLF") {
-			SendEditor(SCI_SETEOLMODE, SC_EOL_CRLF);
-		}
+		SetEol();
 	}
 
 	codePage = props.GetInt("code.page");
+	if (unicodeMode != uni8Bit) {
+		// Override properties file to ensure Unicode displayed.
+		codePage = SC_CP_UTF8;
+	}
 	SendEditor(SCI_SETCODEPAGE, codePage);
+	SendOutput(SCI_SETCODEPAGE, codePage);
 
 	characterSet = props.GetInt("character.set");
 
+#ifdef unix
+	SString localeCType = props.Get("LC_CTYPE");
+	if (localeCType.length())
+		setlocale(LC_CTYPE, localeCType.c_str());
+	else
+		setlocale(LC_CTYPE, "C");
+#endif
+
 	SendEditor(SCI_SETCARETFORE,
 	           ColourOfProperty(props, "caret.fore", ColourDesired(0, 0, 0)));
 
@@ -632,8 +765,7 @@ void SciTEBase::ReadProperties() {
 	SString caretLineBack = props.Get("caret.line.back");
 	if (caretLineBack.length()) {
 		SendEditor(SCI_SETCARETLINEVISIBLE, 1);
-		SendEditor(SCI_SETCARETLINEBACK,
-		           ColourFromString(caretLineBack.c_str()).AsLong());
+		SendEditor(SCI_SETCARETLINEBACK, ColourFromString(caretLineBack));
 	} else {
 		SendEditor(SCI_SETCARETLINEVISIBLE, 0);
 	}
@@ -654,12 +786,19 @@ void SciTEBase::ReadProperties() {
 		SendOutput(SCI_SETCARETPERIOD, caretPeriod.value());
 	}
 
-	int caretStrict = props.GetInt("caret.policy.strict") ? CARET_STRICT : 0;
-	int caretSlop = props.GetInt("caret.policy.slop", 1) ? CARET_SLOP : 0;
-	int caretLines = props.GetInt("caret.policy.lines");
-	int caretXEven = props.GetInt("caret.policy.xeven") ? CARET_XEVEN : 0;
-	int caretXJumps = props.GetInt("caret.policy.xjumps") ? CARET_XJUMPS : 0;
-	SendEditor(SCI_SETCARETPOLICY, caretStrict | caretSlop | caretXEven | caretXJumps, caretLines);
+	int caretSlop = props.GetInt("caret.policy.xslop", 1) ? CARET_SLOP : 0;
+	int caretZone = props.GetInt("caret.policy.width", 50);
+	int caretStrict = props.GetInt("caret.policy.xstrict") ? CARET_STRICT : 0;
+	int caretEven = props.GetInt("caret.policy.xeven", 1) ? CARET_EVEN : 0;
+	int caretJumps = props.GetInt("caret.policy.xjumps") ? CARET_JUMPS : 0;
+	SendEditor(SCI_SETXCARETPOLICY, caretStrict | caretSlop | caretEven | caretJumps, caretZone);
+
+	caretSlop = props.GetInt("caret.policy.yslop", 1) ? CARET_SLOP : 0;
+	caretZone = props.GetInt("caret.policy.lines");
+	caretStrict = props.GetInt("caret.policy.ystrict") ? CARET_STRICT : 0;
+	caretEven = props.GetInt("caret.policy.yeven", 1) ? CARET_EVEN : 0;
+	caretJumps = props.GetInt("caret.policy.yjumps") ? CARET_JUMPS : 0;
+	SendEditor(SCI_SETYCARETPOLICY, caretStrict | caretSlop | caretEven | caretJumps, caretZone);
 
 	int visibleStrict = props.GetInt("visible.policy.strict") ? VISIBLE_STRICT : 0;
 	int visibleSlop = props.GetInt("visible.policy.slop", 1) ? VISIBLE_SLOP : 0;
@@ -673,13 +812,13 @@ void SciTEBase::ReadProperties() {
 
 	SString selFore = props.Get("selection.fore");
 	if (selFore.length()) {
-		SendChildren(SCI_SETSELFORE, 1, ColourFromString(selFore.c_str()).AsLong());
+		SendChildren(SCI_SETSELFORE, 1, ColourFromString(selFore));
 	} else {
 		SendChildren(SCI_SETSELFORE, 0, 0);
 	}
 	SString selBack = props.Get("selection.back");
 	if (selBack.length()) {
-		SendChildren(SCI_SETSELBACK, 1, ColourFromString(selBack.c_str()).AsLong());
+		SendChildren(SCI_SETSELBACK, 1, ColourFromString(selBack));
 	} else {
 		if (selFore.length())
 			SendChildren(SCI_SETSELBACK, 0, 0);
@@ -687,6 +826,32 @@ void SciTEBase::ReadProperties() {
 			SendChildren(SCI_SETSELBACK, 1, ColourDesired(0xC0, 0xC0, 0xC0).AsLong());
 	}
 
+	SString foldColour = props.Get("fold.margin.colour");
+	if (foldColour.length()) {
+		SendChildren(SCI_SETFOLDMARGINCOLOUR, 1, ColourFromString(foldColour));
+	} else {
+		SendChildren(SCI_SETFOLDMARGINCOLOUR, 0, 0);
+	}
+	SString foldHiliteColour = props.Get("fold.margin.highlight.colour");
+	if (foldHiliteColour.length()) {
+		SendChildren(SCI_SETFOLDMARGINHICOLOUR, 1, ColourFromString(foldHiliteColour));
+	} else {
+		SendChildren(SCI_SETFOLDMARGINHICOLOUR, 0, 0);
+	}
+
+	SString whitespaceFore = props.Get("whitespace.fore");
+	if (whitespaceFore.length()) {
+		SendChildren(SCI_SETWHITESPACEFORE, 1, ColourFromString(whitespaceFore));
+	} else {
+		SendChildren(SCI_SETWHITESPACEFORE, 0, 0);
+	}
+	SString whitespaceBack = props.Get("whitespace.back");
+	if (whitespaceBack.length()) {
+		SendChildren(SCI_SETWHITESPACEBACK, 1, ColourFromString(whitespaceBack));
+	} else {
+		SendChildren(SCI_SETWHITESPACEBACK, 0, 0);
+	}
+
 	char bracesStyleKey[200];
 	sprintf(bracesStyleKey, "braces.%s.style", language.c_str());
 	bracesStyle = props.GetInt(bracesStyleKey, 0);
@@ -708,6 +873,14 @@ void SciTEBase::ReadProperties() {
 		autoCompleteStartCharacters = props.GetExpanded("autocomplete.*.start.characters");
 	// "" is a quite reasonable value for this setting
 
+	sprintf(key, "autocomplete.%s.fillups", language.c_str());
+	autoCompleteFillUpCharacters = props.GetExpanded(key);
+	if (autoCompleteFillUpCharacters == "")
+		autoCompleteFillUpCharacters =
+			props.GetExpanded("autocomplete.*.fillups");
+	SendEditorString(SCI_AUTOCSETFILLUPS, 0,
+		autoCompleteFillUpCharacters.c_str());
+
 	sprintf(key, "autocomplete.%s.ignorecase", "*");
 	sval = props.GetNewExpand(key);
 	autoCompleteIgnoreCase = sval == "1";
@@ -758,6 +931,10 @@ void SciTEBase::ReadProperties() {
 		ReadPropertiesInitial();
 	}
 
+	if (useMonoFont) {
+		SetMonoFont();
+	}
+
 	SendEditor(SCI_SETUSEPALETTE, props.GetInt("use.palette"));
 	SendEditor(SCI_SETPRINTMAGNIFICATION, props.GetInt("print.magnification"));
 	SendEditor(SCI_SETPRINTCOLOURMODE, props.GetInt("print.colour.mode"));
@@ -773,38 +950,43 @@ void SciTEBase::ReadProperties() {
 	SendOutput(SCI_SETMARGINRIGHT, 0, blankMarginRight);
 
 	SendEditor(SCI_SETMARGINWIDTHN, 1, margin ? marginWidth : 0);
-	SendEditor(SCI_SETMARGINWIDTHN, 0, lineNumbers ? lineNumbersWidth : 0);
+
+	SString lineMarginProp = props.Get("line.margin.width");
+	lineNumbersWidth = lineMarginProp.value();
+	if (lineNumbersWidth == 0)
+		lineNumbersWidth = lineNumbersWidthDefault;
+	lineNumbersExpand = lineMarginProp.contains('+');
+
+	SetLineNumberWidth();
 
 	bufferedDraw = props.GetInt("buffered.draw", 1);
 	SendEditor(SCI_SETBUFFEREDDRAW, bufferedDraw);
 
+	twoPhaseDraw = props.GetInt("two.phase.draw", 1);
+	SendEditor(SCI_SETTWOPHASEDRAW, twoPhaseDraw);
+
+	SendEditor(SCI_SETLAYOUTCACHE, props.GetInt("cache.layout"));
+	SendOutput(SCI_SETLAYOUTCACHE, props.GetInt("output.cache.layout"));
+
 	bracesCheck = props.GetInt("braces.check");
 	bracesSloppy = props.GetInt("braces.sloppy");
 
+	SendEditor(SCI_SETCHARSDEFAULT);
 	wordCharacters = props.GetNewExpand("word.characters.", fileNameForExtension.c_str());
 	if (wordCharacters.length()) {
 		SendEditorString(SCI_SETWORDCHARS, 0, wordCharacters.c_str());
 	} else {
-		SendEditor(SCI_SETWORDCHARS, 0, 0);
 		wordCharacters = "_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
 	}
 
-	SString useTabs = props.GetNewExpand("use.tab.characters.",
-	                                     fileNameForExtension.c_str());
-	if (useTabs.length())
-		SendEditor(SCI_SETUSETABS, useTabs.value());
-	else
-		SendEditor(SCI_SETUSETABS, props.GetInt("use.tabs", 1));
+	whitespaceCharacters = props.GetNewExpand("whitespace.characters.", fileNameForExtension.c_str());
+	if (whitespaceCharacters.length()) {
+		SendEditorString(SCI_SETWHITESPACECHARS, 0, whitespaceCharacters.c_str());
+	}
 
 	SendEditor(SCI_SETTABINDENTS, props.GetInt("tab.indents", 1));
 	SendEditor(SCI_SETBACKSPACEUNINDENTS, props.GetInt("backspace.unindents", 1));
 
-	int tabSize = props.GetInt("tabsize");
-	if (tabSize) {
-		SendEditor(SCI_SETTABWIDTH, tabSize);
-	}
-	indentSize = props.GetInt("indent.size");
-	SendEditor(SCI_SETINDENT, indentSize);
 	indentOpening = props.GetInt("indent.opening");
 	indentClosing = props.GetInt("indent.closing");
 	indentMaintain = props.GetNewExpand("indent.maintain.", fileNameForExtension.c_str()).value();
@@ -829,22 +1011,39 @@ void SciTEBase::ReadProperties() {
 	preprocCondEnd.Clear();
 	preprocCondEnd.Set(list.c_str());
 
-	matchCase = props.GetInt("find.replace.matchcase");
-	regExp = props.GetInt("find.replace.regexp");
-	unSlash = props.GetInt("find.replace.escapes");
-	wrapFind = props.GetInt("find.replace.wrap", 1);
 	memFiles.AppendList(props.GetNewExpand("find.files"));
 
-	if (props.GetInt("vc.home.key", 1)) {
-		AssignKey(SCK_HOME, 0, SCI_VCHOME);
-		AssignKey(SCK_HOME, SCMOD_SHIFT, SCI_VCHOMEEXTEND);
+	SendEditor(SCI_SETWRAPVISUALFLAGS, props.GetInt("wrap.visual.flags"));
+	SendEditor(SCI_SETWRAPVISUALFLAGSLOCATION, props.GetInt("wrap.visual.flags.location"));
+ 	SendEditor(SCI_SETWRAPSTARTINDENT, props.GetInt("wrap.visual.startindent"));
+
+	if (props.GetInt("wrap.aware.home.end.keys",0)) {
+		if (props.GetInt("vc.home.key", 1)) {
+			AssignKey(SCK_HOME, 0, SCI_VCHOMEWRAP);
+			AssignKey(SCK_HOME, SCMOD_SHIFT, SCI_VCHOMEWRAPEXTEND);
+			AssignKey(SCK_HOME, SCMOD_SHIFT | SCMOD_ALT, SCI_VCHOMERECTEXTEND);
+		} else {
+			AssignKey(SCK_HOME, 0, SCI_HOMEWRAP);
+			AssignKey(SCK_HOME, SCMOD_SHIFT, SCI_HOMEWRAPEXTEND);
+			AssignKey(SCK_HOME, SCMOD_SHIFT | SCMOD_ALT, SCI_HOMERECTEXTEND);
+		}
+		AssignKey(SCK_END, 0, SCI_LINEENDWRAP);
+		AssignKey(SCK_END, SCMOD_SHIFT, SCI_LINEENDWRAPEXTEND);
 	} else {
-		AssignKey(SCK_HOME, 0, SCI_HOME);
-		AssignKey(SCK_HOME, SCMOD_SHIFT, SCI_HOMEEXTEND);
+		if (props.GetInt("vc.home.key", 1)) {
+			AssignKey(SCK_HOME, 0, SCI_VCHOME);
+			AssignKey(SCK_HOME, SCMOD_SHIFT, SCI_VCHOMEEXTEND);
+			AssignKey(SCK_HOME, SCMOD_SHIFT | SCMOD_ALT, SCI_VCHOMERECTEXTEND);
+		} else {
+			AssignKey(SCK_HOME, 0, SCI_HOME);
+			AssignKey(SCK_HOME, SCMOD_SHIFT, SCI_HOMEEXTEND);
+			AssignKey(SCK_HOME, SCMOD_SHIFT | SCMOD_ALT, SCI_HOMERECTEXTEND);
+		}
+		AssignKey(SCK_END, 0, SCI_LINEEND);
+		AssignKey(SCK_END, SCMOD_SHIFT, SCI_LINEENDEXTEND);
 	}
+
 	AssignKey('L', SCMOD_SHIFT | SCMOD_CTRL, SCI_LINEDELETE);
-	SendEditor(SCI_SETHSCROLLBAR, props.GetInt("horizontal.scrollbar", 1));
-	SendOutput(SCI_SETHSCROLLBAR, props.GetInt("output.horizontal.scrollbar", 1));
 
 	scrollOutput = props.GetInt("output.scroll", 1);
 
@@ -859,6 +1058,17 @@ void SciTEBase::ReadProperties() {
 
 	SendEditor(SCI_SETMODEVENTMASK, SC_MOD_CHANGEFOLD);
 
+	if (0==props.GetInt("undo.redo.lazy")) {
+		// Trap for insert/delete notifications (also fired by undo
+		// and redo) so that the buttons can be enabled if needed.
+		SendEditor(SCI_SETMODEVENTMASK, SC_MOD_INSERTTEXT | SC_MOD_DELETETEXT
+			| SC_LASTSTEPINUNDOREDO | SendEditor(SCI_GETMODEVENTMASK, 0));
+
+		//SC_LASTSTEPINUNDOREDO is probably not needed in the mask; it
+		//doesn't seem to fire as an event of its own; just modifies the
+		//insert and delete events.
+	}
+
 	// Create a margin column for the folding symbols
 	SendEditor(SCI_SETMARGINTYPEN, 2, SC_MARGIN_SYMBOL);
 
@@ -920,7 +1130,22 @@ void SciTEBase::ReadProperties() {
 	           ColourOfProperty(props, "bookmark.fore", ColourDesired(0, 0, 0x7f)));
 	SendEditor(SCI_MARKERSETBACK, SciTE_MARKER_BOOKMARK,
 	           ColourOfProperty(props, "bookmark.back", ColourDesired(0x80, 0xff, 0xff)));
-	SendEditor(SCI_MARKERDEFINE, SciTE_MARKER_BOOKMARK, SC_MARK_CIRCLE);
+	SString bookMarkXPM = props.Get("bookmark.pixmap");
+	if (bookMarkXPM.length()) {
+		SendEditorString(SCI_MARKERDEFINEPIXMAP, SciTE_MARKER_BOOKMARK,
+			bookMarkXPM.c_str());
+	} else {
+		SendEditor(SCI_MARKERDEFINE, SciTE_MARKER_BOOKMARK, SC_MARK_CIRCLE);
+	}
+
+	SendEditor(SCI_SETSCROLLWIDTH, props.GetInt("horizontal.scroll.width", 2000));
+	SendOutput(SCI_SETSCROLLWIDTH, props.GetInt("output.horizontal.scroll.width", 2000));
+
+	// Do these last as they force a style refresh
+	SendEditor(SCI_SETHSCROLLBAR, props.GetInt("horizontal.scrollbar", 1));
+	SendOutput(SCI_SETHSCROLLBAR, props.GetInt("output.horizontal.scrollbar", 1));
+
+	SendEditor(SCI_SETENDATLASTLINE, props.GetInt("end.at.last.line", 1));
 
 	if (extender) {
 		extender->Clear();
@@ -928,10 +1153,6 @@ void SciTEBase::ReadProperties() {
 		char defaultDir[MAX_PATH];
 		GetDefaultDirectory(defaultDir, sizeof(defaultDir));
 		char scriptPath[MAX_PATH];
-		if (Exists(defaultDir, "SciTEGlobal.lua", scriptPath)) {
-			// Found fglobal file in global directory
-			extender->Load(scriptPath);
-		}
 
 		// Check for an extension script
 		SString extensionFile = props.GetNewExpand("extension.", fileNameForExtension.c_str());
@@ -958,7 +1179,7 @@ void SciTEBase::ReadProperties() {
 // Properties that are interactively modifiable are only read from the properties file once.
 void SciTEBase::SetPropertiesInitial() {
 	splitVertical = props.GetInt("split.vertical");
-	checkIfOpen = props.GetInt("check.if.already.open");
+	openFilesHere = props.GetInt("check.if.already.open");
 	wrap = props.GetInt("wrap");
 	wrapOutput = props.GetInt("output.wrap");
 	indentationWSVisible = props.GetInt("view.indentation.whitespace", 1);
@@ -966,13 +1187,7 @@ void SciTEBase::SetPropertiesInitial() {
 	tbVisible = props.GetInt("toolbar.visible");
 	tabVisible = props.GetInt("tabbar.visible");
 	tabMultiLine = props.GetInt("tabbar.multiline");
-	lineNumbersWidth = 0;
-	SString linenums = props.Get("line.numbers");
-	if (linenums.length())
-		lineNumbersWidth = linenums.value();
-	lineNumbers = lineNumbersWidth;
-	if (lineNumbersWidth == 0)
-		lineNumbersWidth = lineNumbersWidthDefault;
+	lineNumbers = props.GetInt("line.margin.visible");
 	marginWidth = 0;
 	SString margwidth = props.Get("margin.width");
 	if (margwidth.length())
@@ -984,12 +1199,18 @@ void SciTEBase::SetPropertiesInitial() {
 	foldMargin = foldMarginWidth;
 	if (foldMarginWidth == 0)
 		foldMarginWidth = foldMarginWidthDefault;
+
+	matchCase = props.GetInt("find.replace.matchcase");
+	regExp = props.GetInt("find.replace.regexp");
+	unSlash = props.GetInt("find.replace.escapes");
+	wrapFind = props.GetInt("find.replace.wrap", 1);
 }
 
 SString SciTEBase::LocaliseString(const char *s, bool retainIfNotFound) {
 	SString translation = s;
 	int ellipseIndicator = translation.remove("...");
 	int accessKeyPresent = translation.remove(menuAccessIndicator);
+	translation.lowercase();
 	translation = propsUI.Get(translation.c_str());
 	if (translation.length()) {
 		if (ellipseIndicator)
@@ -1006,12 +1227,14 @@ SString SciTEBase::LocaliseString(const char *s, bool retainIfNotFound) {
 	return s;
 }
 
-SString SciTEBase::LocaliseMessage(const char *s, const char *param0, const char *param1) {
+SString SciTEBase::LocaliseMessage(const char *s, const char *param0, const char *param1, const char *param2) {
 	SString translation = LocaliseString(s);
 	if (param0)
 		translation.substitute("^0", param0);
 	if (param1)
 		translation.substitute("^1", param1);
+	if (param2)
+		translation.substitute("^2", param2);
 	return translation;
 }
 
@@ -1019,11 +1242,15 @@ void SciTEBase::ReadLocalisation() {
 	char propfile[MAX_PATH + 20];
 	char propdir[MAX_PATH + 20];
 	propsUI.Clear();
-	if (GetAbbrevPropertiesFileName(propfile, propdir, sizeof(propfile))) {
-		SString spf(propfile);
-		spf.substitute(propAbbrevFileName, "locale.properties");
+	const char *title = "locale.properties";
+	SString localeProps = props.GetExpanded(title);
+	if (localeProps.length()) {
+		title = localeProps.c_str();
+	}
+	if (GetSciteDefaultHome(propdir, sizeof propdir)
+	    && BuildPath(propfile, propdir, title, sizeof propfile)) {
 		strcat(propdir, pathSepString);
-		propsUI.Read(spf.c_str(), propdir, importFiles, importMax);
+		propsUI.Read(propfile, propdir, importFiles, importMax);
 	}
 	localisationRead = true;
 }
@@ -1032,7 +1259,8 @@ void SciTEBase::ReadPropertiesInitial() {
 	SetPropertiesInitial();
 	int sizeHorizontal = props.GetInt("output.horizontal.size", 0);
 	int sizeVertical = props.GetInt("output.vertical.size", 0);
-	if (splitVertical && sizeVertical > 0 && heightOutput < sizeVertical || sizeHorizontal && heightOutput < sizeHorizontal) {
+	if ((!splitVertical && (sizeVertical > 0) && (heightOutput < sizeVertical)) ||
+		(splitVertical && (sizeHorizontal > 0) && (heightOutput < sizeHorizontal))) {
 		heightOutput = NormaliseSplit(splitVertical ? sizeHorizontal : sizeVertical);
 		SizeSubWindows();
 		Redraw();
@@ -1068,6 +1296,28 @@ void SciTEBase::ReadPropertiesInitial() {
 	}
 	SetLanguageMenu();
 
+	// load the user defined short cut props
+	SString shortCutProp = props.GetNewExpand("user.shortcuts");
+	if ( shortCutProp.length() ) {
+		shortCutItems = 0;
+		for (unsigned int i = 0; i < shortCutProp.length(); i++) {
+			if (shortCutProp[i] == '|')
+				shortCutItems++;
+		}
+		shortCutItems /= 2;
+		shortCutItemList = new ShortcutItem[shortCutItems];
+		shortCutProp.substitute('|', '\0');
+		const char *sShortCutProp = shortCutProp.c_str();
+		for (int item = 0; item < shortCutItems; item++) {
+			shortCutItemList[item].menuKey = sShortCutProp;
+			sShortCutProp += strlen(sShortCutProp) + 1;
+			shortCutItemList[item].menuCommand = sShortCutProp;
+			sShortCutProp += strlen(sShortCutProp) + 1;
+		}
+	}
+	// end load the user defined short cut props
+
+
 #if PLAT_WIN
 
 	if (tabMultiLine) {	// Windows specific!
@@ -1085,6 +1335,30 @@ void SciTEBase::ReadPropertiesInitial() {
 	}
 }
 
+bool SciTEBase::GetDefaultPropertiesFileName(char *pathDefaultProps,
+        char *pathDefaultDir, unsigned int lenPath) {
+	if (!GetSciteDefaultHome(pathDefaultDir, lenPath)) {
+		return false;
+	}
+	return BuildPath(pathDefaultProps, pathDefaultDir, propGlobalFileName, lenPath);
+}
+
+bool SciTEBase::GetAbbrevPropertiesFileName(char *pathAbbrevProps,
+        char *pathDefaultDir, unsigned int lenPath) {
+	if (!GetSciteDefaultHome(pathDefaultDir, lenPath)) {
+		return false;
+	}
+	return BuildPath(pathAbbrevProps, pathDefaultDir, propAbbrevFileName, lenPath);
+}
+
+bool SciTEBase::GetUserPropertiesFileName(char *pathUserProps,
+        char *pathUserDir, unsigned int lenPath) {
+	if (!GetSciteUserHome(pathUserDir, lenPath)) {
+		return false;
+	}
+	return BuildPath(pathUserProps, pathUserDir, propUserFileName, lenPath);
+}
+
 void SciTEBase::OpenProperties(int propsFile) {
 	char propfile[MAX_PATH + 20];
 	char propdir[MAX_PATH + 20];
@@ -1097,18 +1371,169 @@ void SciTEBase::OpenProperties(int propsFile) {
 
 		strcat(propfile, pathSepString);
 		strcat(propfile, propFileName);
-		Open(propfile);
+		Open(propfile, ofQuiet);
 	} else if (propsFile == IDM_OPENUSERPROPERTIES) {
 		if (GetUserPropertiesFileName(propfile, propdir, sizeof(propfile))) {
-			Open(propfile);
+			Open(propfile, ofQuiet);
 		}
 	} else if (propsFile == IDM_OPENABBREVPROPERTIES) {
 		if (GetAbbrevPropertiesFileName(propfile, propdir, sizeof(propfile))) {
-			Open(propfile);
+			Open(propfile, ofQuiet);
 		}
 	} else {	// IDM_OPENGLOBALPROPERTIES
 		if (GetDefaultPropertiesFileName(propfile, propdir, sizeof(propfile))) {
-			Open(propfile);
+			Open(propfile, ofQuiet);
 		}
 	}
 }
+
+// return the int value of the command name passed in.
+int SciTEBase::GetMenuCommandAsInt(SString commandName) {
+	struct {
+		const char *name;
+		int id;
+	} idmTable[] = {
+//++Autogenerated -- run ../scintilla/src/LexGen.py to regenerate
+//**\(\t\t{"\*",\t\*},\n\)
+		{"IDM_MRUFILE",	IDM_MRUFILE},
+		{"IDM_TOOLS",	IDM_TOOLS},
+		{"IDM_BUFFER",	IDM_BUFFER},
+		{"IDM_IMPORT",	IDM_IMPORT},
+		{"IDM_LANGUAGE",	IDM_LANGUAGE},
+		{"IDM_NEW",	IDM_NEW},
+		{"IDM_OPEN",	IDM_OPEN},
+		{"IDM_OPENSELECTED",	IDM_OPENSELECTED},
+		{"IDM_REVERT",	IDM_REVERT},
+		{"IDM_CLOSE",	IDM_CLOSE},
+		{"IDM_SAVE",	IDM_SAVE},
+		{"IDM_SAVEAS",	IDM_SAVEAS},
+		{"IDM_SAVEASHTML",	IDM_SAVEASHTML},
+		{"IDM_SAVEASRTF",	IDM_SAVEASRTF},
+		{"IDM_SAVEASPDF",	IDM_SAVEASPDF},
+		{"IDM_FILER",	IDM_FILER},
+		{"IDM_SAVEASTEX",	IDM_SAVEASTEX},
+		{"IDM_SAVEACOPY",	IDM_SAVEACOPY},
+		{"IDM_SAVEASXML",	IDM_SAVEASXML},
+		{"IDM_MRU_SEP",	IDM_MRU_SEP},
+		{"IDM_PRINTSETUP",	IDM_PRINTSETUP},
+		{"IDM_PRINT",	IDM_PRINT},
+		{"IDM_LOADSESSION",	IDM_LOADSESSION},
+		{"IDM_SAVESESSION",	IDM_SAVESESSION},
+		{"IDM_QUIT",	IDM_QUIT},
+		{"IDM_ENCODING_DEFAULT",	IDM_ENCODING_DEFAULT},
+		{"IDM_ENCODING_UCS2BE",	IDM_ENCODING_UCS2BE},
+		{"IDM_ENCODING_UCS2LE",	IDM_ENCODING_UCS2LE},
+		{"IDM_ENCODING_UTF8",	IDM_ENCODING_UTF8},
+		{"IDM_ENCODING_UCOOKIE",	IDM_ENCODING_UCOOKIE},
+		{"IDM_UNDO",	IDM_UNDO},
+		{"IDM_REDO",	IDM_REDO},
+		{"IDM_CUT",	IDM_CUT},
+		{"IDM_COPY",	IDM_COPY},
+		{"IDM_PASTE",	IDM_PASTE},
+		{"IDM_CLEAR",	IDM_CLEAR},
+		{"IDM_SELECTALL",	IDM_SELECTALL},
+		{"IDM_PASTEANDDOWN",	IDM_PASTEANDDOWN},
+		{"IDM_FIND",	IDM_FIND},
+		{"IDM_FINDNEXT",	IDM_FINDNEXT},
+		{"IDM_FINDNEXTBACK",	IDM_FINDNEXTBACK},
+		{"IDM_FINDNEXTSEL",	IDM_FINDNEXTSEL},
+		{"IDM_FINDNEXTBACKSEL",	IDM_FINDNEXTBACKSEL},
+		{"IDM_FINDINFILES",	IDM_FINDINFILES},
+		{"IDM_REPLACE",	IDM_REPLACE},
+		{"IDM_GOTO",	IDM_GOTO},
+		{"IDM_BOOKMARK_NEXT",	IDM_BOOKMARK_NEXT},
+		{"IDM_BOOKMARK_TOGGLE",	IDM_BOOKMARK_TOGGLE},
+		{"IDM_BOOKMARK_PREV",	IDM_BOOKMARK_PREV},
+		{"IDM_BOOKMARK_CLEARALL",	IDM_BOOKMARK_CLEARALL},
+		{"IDM_MATCHBRACE",	IDM_MATCHBRACE},
+		{"IDM_SELECTTOBRACE",	IDM_SELECTTOBRACE},
+		{"IDM_SHOWCALLTIP",	IDM_SHOWCALLTIP},
+		{"IDM_COMPLETE",	IDM_COMPLETE},
+		{"IDM_COMPLETEWORD",	IDM_COMPLETEWORD},
+		{"IDM_EXPAND",	IDM_EXPAND},
+		{"IDM_TOGGLE_FOLDALL",	IDM_TOGGLE_FOLDALL},
+		{"IDM_UPRCASE",	IDM_UPRCASE},
+		{"IDM_LWRCASE",	IDM_LWRCASE},
+		{"IDM_ABBREV",	IDM_ABBREV},
+		{"IDM_BLOCK_COMMENT",	IDM_BLOCK_COMMENT},
+		{"IDM_STREAM_COMMENT",	IDM_STREAM_COMMENT},
+		{"IDM_COPYASRTF",	IDM_COPYASRTF},
+		{"IDM_BOX_COMMENT",	IDM_BOX_COMMENT},
+		{"IDM_INS_ABBREV",	IDM_INS_ABBREV},
+		{"IDM_JOIN",	IDM_JOIN},
+		{"IDM_SPLIT",	IDM_SPLIT},
+		{"IDM_INCSEARCH",	IDM_INCSEARCH},
+		{"IDM_ENTERSELECTION",	IDM_ENTERSELECTION},
+		{"IDM_PREVMATCHPPC",	IDM_PREVMATCHPPC},
+		{"IDM_SELECTTOPREVMATCHPPC",	IDM_SELECTTOPREVMATCHPPC},
+		{"IDM_NEXTMATCHPPC",	IDM_NEXTMATCHPPC},
+		{"IDM_SELECTTONEXTMATCHPPC",	IDM_SELECTTONEXTMATCHPPC},
+		{"IDM_COMPILE",	IDM_COMPILE},
+		{"IDM_BUILD",	IDM_BUILD},
+		{"IDM_GO",	IDM_GO},
+		{"IDM_STOPEXECUTE",	IDM_STOPEXECUTE},
+		{"IDM_FINISHEDEXECUTE",	IDM_FINISHEDEXECUTE},
+		{"IDM_NEXTMSG",	IDM_NEXTMSG},
+		{"IDM_PREVMSG",	IDM_PREVMSG},
+		{"IDM_MACRO_SEP",	IDM_MACRO_SEP},
+		{"IDM_MACRORECORD",	IDM_MACRORECORD},
+		{"IDM_MACROSTOPRECORD",	IDM_MACROSTOPRECORD},
+		{"IDM_MACROPLAY",	IDM_MACROPLAY},
+		{"IDM_MACROLIST",	IDM_MACROLIST},
+		{"IDM_ACTIVATE",	IDM_ACTIVATE},
+		{"IDM_SRCWIN",	IDM_SRCWIN},
+		{"IDM_RUNWIN",	IDM_RUNWIN},
+		{"IDM_TOOLWIN",	IDM_TOOLWIN},
+		{"IDM_STATUSWIN",	IDM_STATUSWIN},
+		{"IDM_TABWIN",	IDM_TABWIN},
+		{"IDM_SPLITVERTICAL",	IDM_SPLITVERTICAL},
+		{"IDM_VIEWSPACE",	IDM_VIEWSPACE},
+		{"IDM_VIEWEOL",	IDM_VIEWEOL},
+		{"IDM_VIEWGUIDES",	IDM_VIEWGUIDES},
+		{"IDM_SELMARGIN",	IDM_SELMARGIN},
+		{"IDM_FOLDMARGIN",	IDM_FOLDMARGIN},
+		{"IDM_LINENUMBERMARGIN",	IDM_LINENUMBERMARGIN},
+		{"IDM_VIEWTOOLBAR",	IDM_VIEWTOOLBAR},
+		{"IDM_TOGGLEOUTPUT",	IDM_TOGGLEOUTPUT},
+		{"IDM_VIEWTABBAR",	IDM_VIEWTABBAR},
+		{"IDM_VIEWSTATUSBAR",	IDM_VIEWSTATUSBAR},
+		{"IDM_TOGGLEPARAMETERS",	IDM_TOGGLEPARAMETERS},
+		{"IDM_OPENFILESHERE",	IDM_OPENFILESHERE},
+		{"IDM_WRAP",	IDM_WRAP},
+		{"IDM_WRAPOUTPUT",	IDM_WRAPOUTPUT},
+		{"IDM_READONLY",	IDM_READONLY},
+		{"IDM_CLEAROUTPUT",	IDM_CLEAROUTPUT},
+		{"IDM_SWITCHPANE",	IDM_SWITCHPANE},
+		{"IDM_EOL_CRLF",	IDM_EOL_CRLF},
+		{"IDM_EOL_CR",	IDM_EOL_CR},
+		{"IDM_EOL_LF",	IDM_EOL_LF},
+		{"IDM_EOL_CONVERT",	IDM_EOL_CONVERT},
+		{"IDM_TABSIZE",	IDM_TABSIZE},
+		{"IDM_MONOFONT",	IDM_MONOFONT},
+		{"IDM_OPENLOCALPROPERTIES",	IDM_OPENLOCALPROPERTIES},
+		{"IDM_OPENUSERPROPERTIES",	IDM_OPENUSERPROPERTIES},
+		{"IDM_OPENGLOBALPROPERTIES",	IDM_OPENGLOBALPROPERTIES},
+		{"IDM_OPENABBREVPROPERTIES",	IDM_OPENABBREVPROPERTIES},
+		{"IDM_PREVFILE",	IDM_PREVFILE},
+		{"IDM_NEXTFILE",	IDM_NEXTFILE},
+		{"IDM_CLOSEALL",	IDM_CLOSEALL},
+		{"IDM_SAVEALL",	IDM_SAVEALL},
+		{"IDM_BUFFERSEP",	IDM_BUFFERSEP},
+		{"IDM_HELP",	IDM_HELP},
+		{"IDM_ABOUT",	IDM_ABOUT},
+		{"IDM_HELP_SCITE",	IDM_HELP_SCITE},
+		{"IDM_ONTOP",	IDM_ONTOP},
+		{"IDM_FULLSCREEN",	IDM_FULLSCREEN},
+
+//--Autogenerated -- end of automatically generated section
+		{0, 0},
+	};
+
+	for (int i=0; idmTable[i].id; i++) {
+		if (commandName == idmTable[i].name) {
+			return idmTable[i].id;
+		}
+	}
+	// Otherwise we might have entered a number as command to access a "SCI_" command
+	return commandName.value();
+}
