@@ -2,7 +2,7 @@
 /** @file SciTEWinBar.cxx
  ** Bar and menu code for the Windows version of the editor.
  **/
-// Copyright 1998-2002 by Neil Hodgson <neilh@scintilla.org>
+// Copyright 1998-2003 by Neil Hodgson <neilh@scintilla.org>
 // The License.txt file describes the conditions under which this software may be distributed.
 
 #include "SciTEWin.h"
@@ -11,24 +11,10 @@
  * Update the status bar text.
  */
 void SciTEWin::SetStatusBarText(const char *s) {
-	::SendMessage(reinterpret_cast<HWND>(wStatusBar.GetID()), SB_SETTEXT, 1, reinterpret_cast<LPARAM>(s));
+	::SendMessage(reinterpret_cast<HWND>(wStatusBar.GetID()),
+	              SB_SETTEXT, 0, reinterpret_cast<LPARAM>(s));
 }
 
-#ifdef __MINGW_H	
-// Mingw headers do not have NMMOUSE, TTN_GETDISPINFO or NMTTDISPINFO
-struct NMMOUSE {
-	NMHDR hdr;
-	DWORD dwItemSpec;
-	// Other fields
-};
-#define TTN_GETDISPINFO TTN_FIRST
-struct NMTTDISPINFO {
-	NMHDR hdr;
-	LPSTR lpszText;
-	// Other fields
-};
-#endif
-
 /**
  * Manage Windows specific notifications.
  */
@@ -36,7 +22,6 @@ void SciTEWin::Notify(SCNotification *notification) {
 	switch (notification->nmhdr.code) {
 	case TCN_SELCHANGE:
 		// Change of tab
-
 		if (notification->nmhdr.idFrom == IDM_TABWIN) {
 			int index = Platform::SendScintilla(wTabBar.GetID(), TCM_GETCURSEL, (WPARAM)0, (LPARAM)0);
 			SetDocumentAt(index);
@@ -50,16 +35,13 @@ void SciTEWin::Notify(SCNotification *notification) {
 			// Click on the status bar
 			NMMOUSE *pNMMouse = (NMMOUSE *)notification;
 			switch (pNMMouse->dwItemSpec) {
-			case 1:		/* Display of status */
+			case 0: 		/* Display of status */
 				sbNum++;
 				if (sbNum > props.GetInt("statusbar.number")) {
 					sbNum = 1;
 				}
 				UpdateStatusBar(true);
 				break;
-			case 2:		/* Refresh button */
-				UpdateStatusBar(true);
-				break;
 			default:
 				break;
 			}
@@ -69,8 +51,8 @@ void SciTEWin::Notify(SCNotification *notification) {
 	case TTN_GETDISPINFO:
 		// Ask for tooltip text
 		{
-			static char ttt[MAX_PATH];
-			const char *ttext=0;
+			static char ttt[MAX_PATH*2 + 1];
+			const char *ttext = 0;
 			NMTTDISPINFO *pDispInfo = (NMTTDISPINFO *)notification;
 			// Toolbar tooltips
 			switch (notification->nmhdr.idFrom) {
@@ -122,18 +104,29 @@ void SciTEWin::Notify(SCNotification *notification) {
 			case IDM_MACROPLAY:
 				ttext = "Run Macro";
 				break;
-			default:
-				{
+			default: {
 					// notification->nmhdr.idFrom appears to be the buffer number for tabbar tooltips
 					Point ptCursor;
 					::GetCursorPos(reinterpret_cast<POINT *>(&ptCursor));
 					Point ptClient = ptCursor;
-					::ScreenToClient(reinterpret_cast<HWND>(wTabBar.GetID()), 
-						reinterpret_cast<POINT *>(&ptClient));
+					::ScreenToClient(reinterpret_cast<HWND>(wTabBar.GetID()),
+					                 reinterpret_cast<POINT *>(&ptClient));
 					TCHITTESTINFO info;
-					info.pt.x = ptClient.x; info.pt.y = ptClient.y;
+					info.pt.x = ptClient.x;
+					info.pt.y = ptClient.y;
 					int index = Platform::SendScintilla(wTabBar.GetID(), TCM_HITTEST, (WPARAM)0, (LPARAM) & info);
-					pDispInfo->lpszText = const_cast<char *>(buffers.buffers[index].FullPath());
+					if (index >= 0) {
+						SString path = buffers.buffers[index].FullPath();
+						// Handle '&' characters in path, since they are interpreted in
+						// tooltips.
+						int amp = 0;
+						while ((amp = path.search("&", amp)) >= 0) {
+							path.insert(amp, "&");
+							amp += 2;
+						}
+						strcpy(ttt, path.c_str());
+						pDispInfo->lpszText = const_cast<char *>(ttt);
+					}
 				}
 				break;
 			}
@@ -147,20 +140,20 @@ void SciTEWin::Notify(SCNotification *notification) {
 
 	case SCN_CHARADDED:
 		if ((notification->nmhdr.idFrom == IDM_RUNWIN) &&
-			executing &&
-			hWriteSubProcess) {
+		        executing &&
+		        hWriteSubProcess) {
 			char chToWrite = static_cast<char>(notification->ch);
 			if (chToWrite != '\r') {
 				DWORD bytesWrote = 0;
 				::WriteFile(hWriteSubProcess, &chToWrite,
-									   1, &bytesWrote, NULL);
+				            1, &bytesWrote, NULL);
 			}
 		} else {
 			SciTEBase::Notify(notification);
 		}
 		break;
-	
-	default:    	// Scintilla notification, use default treatment
+
+	default:     	// Scintilla notification, use default treatment
 		SciTEBase::Notify(notification);
 		break;
 	}
@@ -215,8 +208,12 @@ void SciTEWin::SizeSubWindows() {
 	}
 
 	if (showTab) {
-		int tabNb = ::SendMessage(reinterpret_cast<HWND>(wTabBar.GetID()), TCM_GETROWCOUNT, 0, 0);
-		visHeightTab = tabNb * heightTab;
+		wTabBar.SetPosition(PRectangle(
+			rcClient.left, rcClient.top + visHeightTools,
+			rcClient.right, rcClient.top + heightTab + visHeightTools));
+		int tabNb = ::SendMessage(reinterpret_cast<HWND>(
+			wTabBar.GetID()), TCM_GETROWCOUNT, 0, 0);
+		visHeightTab = ((tabNb - 1) * (heightTab - 6)) + heightTab;
 	} else {
 		visHeightTab = 0;
 	}
@@ -230,7 +227,7 @@ void SciTEWin::SizeSubWindows() {
 	}
 	if (tbVisible) {
 		wToolBar.SetPosition(PRectangle(
-		                         rcClient.left, rcClient.top, rcClient.Width(), visHeightTools));
+		                         rcClient.left, rcClient.top, rcClient.right, visHeightTools));
 		wToolBar.Show(true);
 	} else {
 		wToolBar.Show(false);
@@ -240,7 +237,7 @@ void SciTEWin::SizeSubWindows() {
 	if (showTab) {
 		wTabBar.SetPosition(PRectangle(
 		                        rcClient.left, rcClient.top + visHeightTools,
-		                        rcClient.Width(), visHeightTab + visHeightTools));
+		                        rcClient.right, rcClient.top + visHeightTab + visHeightTools));
 		wTabBar.Show(true);
 	} else {
 		wTabBar.Show(false);
@@ -249,29 +246,10 @@ void SciTEWin::SizeSubWindows() {
 		                        rcClient.Width(), 1));
 	}
 	if (sbVisible) {
-		int spw = props.GetInt("statusbar.size");
-		if (spw <= 0) {
-			spw = statusPosWidth;
-		}
-		int refreshWidth = 70;
-		int startLineNum = rcClient.Width() - spw - refreshWidth;
-		if (startLineNum < 0)
-			startLineNum = 0;
-		int widths[] = { startLineNum, rcClient.Width() - refreshWidth, rcClient.Width() };
-		// Perhaps we can define a syntax to create more parts,
-		// but it is probably an overkill for a marginal feature
-		::SendMessage(reinterpret_cast<HWND>(wStatusBar.GetID()), 
-		              SB_SETPARTS, 3,
-		              reinterpret_cast<LPARAM>(widths));
-		::SendMessage(reinterpret_cast<HWND>(wStatusBar.GetID()), 
-		              SB_SETTEXT, 0 | SBT_NOBORDERS,
-		              reinterpret_cast<LPARAM>(""));
-		::SendMessage(reinterpret_cast<HWND>(wStatusBar.GetID()), 
-		              SB_SETTEXT, 2 | SBT_POPOUT,
-		              reinterpret_cast<LPARAM>("Refresh"));
-		wStatusBar.SetPosition(PRectangle(rcClient.left,
-		                                  rcClient.top + visHeightTools + visHeightTab + visHeightEditor,
-		                                  rcClient.Width(), visHeightStatus));
+		wStatusBar.SetPosition(PRectangle(
+		                           rcClient.left, rcClient.top + visHeightTools + visHeightTab + visHeightEditor,
+		                           rcClient.right,
+		                           rcClient.top + visHeightTools + visHeightTab + visHeightEditor + visHeightStatus));
 		wStatusBar.Show(true);
 	} else {
 		wStatusBar.Show(false);
@@ -279,23 +257,42 @@ void SciTEWin::SizeSubWindows() {
 		                           rcClient.left, rcClient.top - 2, rcClient.Width(), 1));
 	}
 
-	wContent.SetPosition(PRectangle(rcClient.left, visHeightTools + visHeightTab, rcClient.Width(),
-	                                visHeightTools + visHeightTab + visHeightEditor));
+	wContent.SetPosition(PRectangle(
+	                         rcClient.left, rcClient.top + visHeightTab + visHeightTools,
+	                         rcClient.right,
+	                         rcClient.top + visHeightTab + visHeightTools + visHeightEditor));
 	SizeContentWindows();
 	//::SendMessage(MainHWND(), WM_SETREDRAW, true, 0);
 	//::RedrawWindow(MainHWND(), NULL, NULL, RDW_INVALIDATE | RDW_ALLCHILDREN);
 }
 
+
+
+// Keymod param is interpreted using the same notation (and much the same 
+// code) as KeyMatch uses in SciTEWin.cxx.
+
+
+
 void SciTEWin::SetMenuItem(int menuNumber, int position, int itemID,
                            const char *text, const char *mnemonic) {
 	// On Windows the menu items are modified if they already exist or are created
 	HMENU hmenuBar = ::GetMenu(MainHWND());
 	HMENU hmenu = ::GetSubMenu(hmenuBar, menuNumber);
 	SString sTextMnemonic = text;
-	if (mnemonic) {
-		sTextMnemonic += "\t";
-		sTextMnemonic += mnemonic;
+	long keycode;
+	if (mnemonic && *mnemonic) {
+		keycode = SciTEKeys::ParseKeyCode(mnemonic);
+		if (keycode) {
+			sTextMnemonic += "\t";
+		sTextMnemonic += LocaliseAccelerator(mnemonic, itemID);
 	}
+		// the keycode could be used to make a custom accelerator table
+		// but for now, the menu's item data is used instead for command
+		// tools, and for other menu entries it is just discarded.
+	} else {
+		keycode = 0; //I don't think this is needed in ANSI C++.
+	}
+
 	if (::GetMenuState(hmenu, itemID, MF_BYCOMMAND) == 0xffffffff) {
 		if (text[0])
 			::InsertMenu(hmenu, position, MF_BYPOSITION, itemID, sTextMnemonic.c_str());
@@ -304,6 +301,16 @@ void SciTEWin::SetMenuItem(int menuNumber, int position, int itemID,
 	} else {
 		::ModifyMenu(hmenu, position, MF_BYCOMMAND, itemID, sTextMnemonic.c_str());
 	}
+
+	if (itemID >= IDM_TOOLS && itemID < IDM_TOOLS + toolMax) {
+		// Stow the keycode for later retrieval.
+		// Do this even if 0, in case the menu already existed (e.g. ModifyMenu)
+		MENUITEMINFO mii;
+		mii.cbSize = sizeof(MENUITEMINFO);
+		mii.fMask = MIIM_DATA;
+		mii.dwItemData = reinterpret_cast<DWORD&>(keycode);
+		::SetMenuItemInfo(hmenu, itemID, FALSE, &mii);
+	}
 }
 
 void SciTEWin::DestroyMenuItem(int menuNumber, int itemID) {
@@ -325,8 +332,10 @@ void SciTEWin::CheckAMenuItem(int wIDCheckItem, bool val) {
 }
 
 void EnableButton(HWND wTools, int id, bool enable) {
-	::SendMessage(wTools, TB_ENABLEBUTTON, id,
+	if (wTools) {
+		::SendMessage(wTools, TB_ENABLEBUTTON, id,
 	              Platform::LongFromTwoShorts(static_cast<short>(enable ? TRUE : FALSE), 0));
+	}
 }
 
 void SciTEWin::EnableAMenuItem(int wIDCheckItem, bool val) {
@@ -341,18 +350,122 @@ void SciTEWin::CheckMenus() {
 	SciTEBase::CheckMenus();
 	CheckMenuRadioItem(GetMenu(MainHWND()), IDM_EOL_CRLF, IDM_EOL_LF,
 	                   SendEditor(SCI_GETEOLMODE) - SC_EOL_CRLF + IDM_EOL_CRLF, 0);
+	CheckMenuRadioItem(GetMenu(MainHWND()), IDM_ENCODING_DEFAULT, IDM_ENCODING_UCOOKIE,
+	                   unicodeMode + IDM_ENCODING_DEFAULT, 0);
+}
+
+void SciTEWin::MakeAccelerator(SString sAccelerator, ACCEL &Accel) {
+	SString s = sAccelerator;
+
+	if (s.contains("null")) {
+		Accel.key = 0;
+		return ;
+	}
+
+	if (s.contains("Ctrl+")) {
+		Accel.fVirt |= FCONTROL;
+		s.remove("Ctrl+");
+	}
+	if (s.contains("Shift+")) {
+		Accel.fVirt |= FSHIFT;
+		s.remove("Shift+");
+	}
+	if (s.contains("Alt+")) {
+		Accel.fVirt |= FALT;
+		s.remove("Alt+");
+	}
+	if (s.length() == 1) {
+		Accel.key = s[0];
+		Accel.fVirt |= FVIRTKEY;
+	} else if ((s.length() > 1) && (s[0] == 'F') && (isdigit(s[1]))) {
+		s.remove("F");
+		int keyNum = s.value();
+		Accel.key = static_cast<WORD>(keyNum + VK_F1 - 1);
+		Accel.fVirt |= FVIRTKEY;
+	} else if (s.contains("Del")) {
+		Accel.key = VK_DELETE;
+		Accel.fVirt |= FVIRTKEY;
+	} else if (s.contains("Space")) {
+		Accel.key = VK_SPACE;
+		Accel.fVirt |= FVIRTKEY;
+	} else if (s.contains("Enter")) {
+		Accel.key = VK_RETURN;
+		Accel.fVirt |= FVIRTKEY;
+	} else if (s.contains("Back")) {
+		Accel.key = VK_BACK;
+		Accel.fVirt |= FVIRTKEY;
+	} else if (s.contains("Tab")) {
+		Accel.key = VK_TAB;
+		Accel.fVirt |= FVIRTKEY;
+	} else if (s.contains("Num")) {
+		Accel.fVirt |= FVIRTKEY;
+		s.remove("Num");
+		if (isdigit(s[0])) {
+			int keyNum = s.value();
+			Accel.key = static_cast<WORD>(keyNum + VK_NUMPAD0);
+		} else {
+			switch (s[0]) {
+			case '*':
+				Accel.key = VK_MULTIPLY;
+				break;
+			case '+':
+				Accel.key = VK_ADD;
+				break;
+			case '-':
+				Accel.key = VK_SUBTRACT;
+				break;
+			case '/':
+				Accel.key = VK_DIVIDE;
+				break;
+			default:
+				Accel.key = 0;
+				break;
+			}
+		}
+	}
+}
+
+//SString SciTEWin::LocaliseAccelerator(const char *pAccelerator, int cmd) {
+SString SciTEWin::LocaliseAccelerator(const char *pAccelerator, int) {
+#ifdef LOCALISE_ACCELERATORS_WORKED
+	SString translation = LocaliseString(pAccelerator, true);
+	int AccelCount = ::CopyAcceleratorTable(hAccTable, NULL, 0);
+	ACCEL *AccelTable = new ACCEL[AccelCount];
+	::CopyAcceleratorTable(hAccTable, AccelTable, AccelCount);
+	for (int i = 0; i < AccelCount; i++) {
+		if (AccelTable[i].cmd == cmd) {
+			MakeAccelerator(translation, AccelTable[i]);
+		}
+	}
+
+	::DestroyAcceleratorTable(hAccTable);
+	hAccTable = ::CreateAcceleratorTable(AccelTable, AccelCount);
+	delete []AccelTable;
+
+	if (translation.contains("null")) {
+		translation.clear();
+	}
+
+	return translation;
+#else
+	return pAccelerator;
+#endif
 }
 
 void SciTEWin::LocaliseMenu(HMENU hmenu) {
-	for (int i=0; i<=::GetMenuItemCount(hmenu); i++) {
+	for (int i = 0; i <= ::GetMenuItemCount(hmenu); i++) {
 		char buff[200];
+		buff[0] = '\0';
 		MENUITEMINFO mii;
 		memset(&mii, 0, sizeof(mii));
-		mii.cbSize = sizeof(mii);
-		mii.fMask = MIIM_CHECKMARKS | MIIM_DATA | MIIM_ID | 
-			MIIM_STATE | MIIM_SUBMENU | MIIM_TYPE;
+		// Explicitly use the struct size for NT 4 as otherwise
+		// GetMenuItemInfo will fail on NT 4.
+		//mii.cbSize = sizeof(mii);
+		mii.cbSize = 44;
+		mii.fMask = MIIM_CHECKMARKS | MIIM_DATA | MIIM_ID |
+		            MIIM_STATE | MIIM_SUBMENU | MIIM_TYPE;
 		mii.dwTypeData = buff;
-		mii.cch = sizeof(buff)-1;
+		mii.cch = sizeof(buff) - 1;
 		if (::GetMenuItemInfo(hmenu, i, TRUE, &mii)) {
 			if (mii.hSubMenu) {
 				LocaliseMenu(mii.hSubMenu);
@@ -361,7 +474,7 @@ void SciTEWin::LocaliseMenu(HMENU hmenu) {
 				if (mii.dwTypeData) {
 					SString text(mii.dwTypeData);
 					SString accel(mii.dwTypeData);
-					int len = text.length();
+					size_t len = text.length();
 					int tab = text.search("\t");
 					if (tab != -1) {
 						text.remove(tab, len - tab);
@@ -369,10 +482,10 @@ void SciTEWin::LocaliseMenu(HMENU hmenu) {
 					} else {
 						accel = "";
 					}
-					text = LocaliseString(text.c_str(), false);
+					text = LocaliseString(text.c_str(), true);
 					if (text.length()) {
 						text += "\t";
-						text += accel;
+						text += LocaliseAccelerator(accel.c_str(), mii.wID);
 						mii.dwTypeData = const_cast<char *>(text.c_str());
 						::SetMenuItemInfo(hmenu, i, TRUE, &mii);
 					}
@@ -387,6 +500,22 @@ void SciTEWin::LocaliseMenus() {
 	::DrawMenuBar(MainHWND());
 }
 
+void SciTEWin::LocaliseAccelerators() {
+	LocaliseAccelerator("Alt+1", IDM_BUFFER + 0);
+	LocaliseAccelerator("Alt+2", IDM_BUFFER + 1);
+	LocaliseAccelerator("Alt+3", IDM_BUFFER + 2);
+	LocaliseAccelerator("Alt+4", IDM_BUFFER + 3);
+	LocaliseAccelerator("Alt+5", IDM_BUFFER + 4);
+	LocaliseAccelerator("Alt+6", IDM_BUFFER + 5);
+	LocaliseAccelerator("Alt+7", IDM_BUFFER + 6);
+	LocaliseAccelerator("Alt+8", IDM_BUFFER + 7);
+	LocaliseAccelerator("Alt+9", IDM_BUFFER + 8);
+	LocaliseAccelerator("Alt+0", IDM_BUFFER + 9);
+
+	// todo read keymap from cfg
+	// AssignKey('Y', SCMOD_CTRL, SCI_LINECUT);
+}
+
 void SciTEWin::LocaliseControl(HWND w) {
 	char wtext[200];
 	if (::GetWindowText(w, wtext, sizeof(wtext))) {
@@ -401,7 +530,7 @@ void SciTEWin::LocaliseDialog(HWND wDialog) {
 	HWND wChild = ::GetWindow(wDialog, GW_CHILD);
 	while (wChild) {
 		LocaliseControl(wChild);
-		wChild = GetWindow(wChild, GW_HWNDNEXT);
+		wChild = ::GetWindow(wChild, GW_HWNDNEXT);
 	}
 }
 
@@ -421,24 +550,24 @@ struct BarButton {
 };
 
 static BarButton bbs[] = {
-    { -1, 0 },
-    { STD_FILENEW, IDM_NEW },
+    { -1,           0 },
+    { STD_FILENEW,  IDM_NEW },
     { STD_FILEOPEN, IDM_OPEN },
     { STD_FILESAVE, IDM_SAVE },
-    { 0, IDM_CLOSE },
-    { -1, 0 },
-    { STD_PRINT, IDM_PRINT },
-    { -1, 0 },
-    { STD_CUT, IDM_CUT },
-    { STD_COPY, IDM_COPY },
-    { STD_PASTE, IDM_PASTE },
-    { STD_DELETE, IDM_CLEAR },
-    { -1, 0 },
-    { STD_UNDO, IDM_UNDO },
-    { STD_REDOW, IDM_REDO },
-    { -1, 0 },
-    { STD_FIND, IDM_FIND },
-    { STD_REPLACE, IDM_REPLACE },
+    { 0,            IDM_CLOSE },
+    { -1,           0 },
+    { STD_PRINT,    IDM_PRINT },
+    { -1,           0 },
+    { STD_CUT,      IDM_CUT },
+    { STD_COPY,     IDM_COPY },
+    { STD_PASTE,    IDM_PASTE },
+    { STD_DELETE,   IDM_CLEAR },
+    { -1,           0 },
+    { STD_UNDO,     IDM_UNDO },
+    { STD_REDOW,    IDM_REDO },
+    { -1,           0 },
+    { STD_FIND,     IDM_FIND },
+    { STD_REPLACE,  IDM_REPLACE },
 };
 
 /**
@@ -474,8 +603,8 @@ void SciTEWin::Creation() {
 		exit(FALSE);
 	fnEditor = reinterpret_cast<SciFnDirect>(::SendMessage(
 	               reinterpret_cast<HWND>(wEditor.GetID()), SCI_GETDIRECTFUNCTION, 0, 0));
-	ptrEditor = ::SendMessage(reinterpret_cast<HWND>(wEditor.GetID()), 
-		SCI_GETDIRECTPOINTER, 0, 0);
+	ptrEditor = ::SendMessage(reinterpret_cast<HWND>(wEditor.GetID()),
+	                          SCI_GETDIRECTPOINTER, 0, 0);
 	wEditor.Show();
 	SendEditor(SCI_USEPOPUP, 0);
 	WindowSetFocus(wEditor);
@@ -495,8 +624,8 @@ void SciTEWin::Creation() {
 		exit(FALSE);
 	fnOutput = reinterpret_cast<SciFnDirect>(::SendMessage(
 	               reinterpret_cast<HWND>(wOutput.GetID()), SCI_GETDIRECTFUNCTION, 0, 0));
-	ptrOutput = ::SendMessage(reinterpret_cast<HWND>(wOutput.GetID()), 
-		SCI_GETDIRECTPOINTER, 0, 0);
+	ptrOutput = ::SendMessage(reinterpret_cast<HWND>(wOutput.GetID()),
+	                          SCI_GETDIRECTPOINTER, 0, 0);
 	wOutput.Show();
 	// No selection margin on output window
 	SendOutput(SCI_SETMARGINWIDTHN, 1, 0);
@@ -509,8 +638,7 @@ void SciTEWin::Creation() {
 	               TOOLBARCLASSNAME,
 	               "",
 	               WS_CHILD | WS_CLIPCHILDREN | WS_CLIPSIBLINGS |
-	               CCS_ADJUSTABLE |
-	               TBSTYLE_FLAT | TBSTYLE_TOOLTIPS,
+	               TBSTYLE_FLAT | TBSTYLE_TOOLTIPS | CCS_NORESIZE,
 	               0, 0,
 	               100, heightTools,
 	               MainHWND(),
@@ -518,12 +646,12 @@ void SciTEWin::Creation() {
 	               hInstance,
 	               0);
 	wToolBar = hwndToolBar;
-	
+
 	::SendMessage(hwndToolBar, TB_BUTTONSTRUCTSIZE, sizeof(TBBUTTON), 0);
 	::SendMessage(hwndToolBar, TB_LOADIMAGES, IDB_STD_SMALL_COLOR,
 	              reinterpret_cast<LPARAM>(HINST_COMMCTRL));
 
-	TBADDBITMAP addbmp= { hInstance, IDR_CLOSEFILE };
+	TBADDBITMAP addbmp = { hInstance, IDR_CLOSEFILE };
 	::SendMessage(hwndToolBar, TB_ADDBITMAP, 1, (LPARAM)&addbmp);
 
 	TBBUTTON tbb[ELEMENTS(bbs)];
@@ -543,7 +671,6 @@ void SciTEWin::Creation() {
 	}
 
 	::SendMessage(hwndToolBar, TB_ADDBUTTONS, ELEMENTS(bbs), reinterpret_cast<LPARAM>(tbb));
-	::SendMessage(hwndToolBar, TB_AUTOSIZE, 0, 0);
 
 	wToolBar.Show();
 
@@ -574,7 +701,7 @@ void SciTEWin::Creation() {
 	                         "Ms Sans Serif");
 	::SendMessage(reinterpret_cast<HWND>(wTabBar.GetID()),
 	              WM_SETFONT,
-	              reinterpret_cast<WPARAM>(fontTabs),     // handle to font
+	              reinterpret_cast<WPARAM>(fontTabs),      // handle to font
 	              0);    // redraw option
 
 	wTabBar.Show();
@@ -591,5 +718,11 @@ void SciTEWin::Creation() {
 	                 hInstance,
 	                 0);
 	wStatusBar.Show();
+	int widths[] = { 4000 };
+	// Perhaps we can define a syntax to create more parts,
+	// but it is probably an overkill for a marginal feature
+	::SendMessage(reinterpret_cast<HWND>(wStatusBar.GetID()),
+	              SB_SETPARTS, 1,
+	              reinterpret_cast<LPARAM>(widths));
 }
 
